# vim: set fileencoding=utf-8
# pylint: disable=C0103

"""
Wrappers for gdal and rasterio.

Copyright (C) 2018, Carlo de Franchis <carlo.de-franchis@ens-cachan.fr>
"""

from __future__ import print_function
import os
import re
import shutil
import argparse
import datetime
import subprocess
import tempfile
import numpy as np
import utm
import traceback
import warnings
import sys
import geojson
import requests
import shapely.geometry
import rasterio
import pyproj

from tsd import rpc_model


warnings.filterwarnings("ignore",
                        category=rasterio.errors.NotGeoreferencedWarning)


def download(from_url, to_file, auth=('', '')):
    """
    Download a file from an url to a file.
    """
    to_file = os.path.abspath(os.path.expanduser(to_file))
    os.makedirs(os.path.dirname(to_file), exist_ok=True)
    response = requests.get(from_url, stream=True, auth=auth)
    with open(to_file, 'wb') as handle:
        for data in response.iter_content():
            handle.write(data)


def valid_datetime(s):
    """
    Check if a string is a well-formatted datetime.
    """
    try:
        return datetime.datetime.strptime(s, "%Y-%m-%d")
    except ValueError:
        raise argparse.ArgumentTypeError("Invalid date: '{}'".format(s))


def valid_date(s):
    """
    Check if a string is a well-formatted date.
    """
    try:
        return datetime.datetime.strptime(s, "%Y-%m-%d").date()
    except ValueError:
        raise argparse.ArgumentTypeError("Invalid date: '{}'".format(s))


def valid_lon(s):
    """
    Check if a string is a well-formatted longitude.

    Args:
        s (str): longitude expressed as a decimal floating number (e.g.
            5.7431861) or as degrees, minutes and seconds (e.g. 5d44'35.47"E)
    """
    try:
        return float(s)
    except ValueError:
        regex = r"(\d+)d(\d+)'([\d.]+)\"([WE])"
        m = re.match(regex, s)
        if m is None:
            raise argparse.ArgumentTypeError("Invalid longitude: '{}'".format(s))
        else:
            x = m.groups()
            lon = int(x[0]) + float(x[1]) / 60 + float(x[2]) / 3600
            if x[3] == 'W':
                lon *= -1
            return lon


def valid_lat(s):
    """
    Check if a string is a well-formatted latitude.

    Args:
        s (str): longitude expressed as a decimal floating number (e.g.
            -49.359053) or as degrees, minutes and seconds (e.g. 49d21'32.59"S)
    """
    try:
        return float(s)
    except ValueError:
        regex = r"(\d+)d(\d+)'([\d.]+)\"([NS])"
        m = re.match(regex, s)
        if m is None:
            raise argparse.ArgumentTypeError("Invalid latitude: '{}'".format(s))
        else:
            x = m.groups()
            lat = int(x[0]) + float(x[1]) / 60 + float(x[2]) / 3600
            if x[3] == 'S':
                lat *= -1
            return lat


def valid_geojson(filepath):
    """
    Check if a file contains valid geojson.
    """
    with open(filepath, 'r') as f:
        geo = geojson.load(f)
    if type(geo) == geojson.geometry.Polygon:
        return geo
    if type(geo) == geojson.feature.FeatureCollection:
        p = geo['features'][0]['geometry']
        if type(p) == geojson.geometry.Polygon:
            return p
    raise argparse.ArgumentTypeError('Invalid geojson: only polygons are supported')


def geojson_geometry_object(lat, lon, w, h):
    """
    """
    return geojson.Polygon([lonlat_rectangle_centered_at(lon, lat, w, h)])


def is_valid(f):
    """
    Check if a file is valid readable image according to rasterio.

    Args:
        f (str): path to a file

    Return:
        boolean telling wether or not the file is a valid image
    """
    try:
        a = rasterio.open(f, 'r')
        a.close()
        return True
    except rasterio.RasterioIOError:
        return False


def tmpfile(ext=''):
    """
    Creates a temporary file.

    Args:
        ext: desired file extension. The dot has to be included.

    Returns:
        absolute path to the created file
    """
    fd, out = tempfile.mkstemp(suffix=ext)
    os.close(fd)           # http://www.logilab.org/blogentry/17873
    return out


def pixel_size(path):
    """
    Read the resolution (in meters per pixel) of a GeoTIFF image.

    Args:
        path (string): path to a GeoTIFF image file

    Return:
        rx, ry (tuple): two floats giving the horizontal and vertical pixel size
    """
    with rasterio.open(path, 'r') as f:
        return f.res


def set_geotif_metadata_items(path, tags={}):
    """
    Append key, value pairs to the GDAL "metadata" tag of a GeoTIFF file.

    Args:
        path (str): path to a GeoTIFF file
        tags (dict): key, value pairs to be added to the "metadata" tag
    """
    with rasterio.open(path, 'r+') as dst:
        dst.update_tags(**tags)


def geotiff_utm_zone(path):
    """
    Read the UTM zone of a GeoTIFF image.

    Args:
        path (string): path to a GeoTIFF image file

    Return:
        int between 1 and 60 identifying the UTM zone
    """
    with rasterio.open(path, 'r') as f:
        return int(f.crs['init'][-2:])


def gdal_translate_version():
    """
    """
    v = subprocess.check_output(['gdal_translate', '--version'])
    return v.decode().split()[1].split(',')[0]


def inplace_utm_reprojection_with_gdalwarp(src, utm_zone, ulx, uly, lrx, lry):
    """
    """
    if geotiff_utm_zone(src) != utm_zone:

        # hack to allow the output to overwrite the input
        fd, dst = tempfile.mkstemp(suffix='.tif', dir=os.path.dirname(src))
        os.close(fd)

        cmd = ['gdalwarp', '-t_srs', '+proj=utm +zone={}'.format(utm_zone),
               '-te', str(ulx), str(lry), str(lrx), str(uly),  # xmin ymin xmax ymax
               '-overwrite', src, dst]
        print(' '.join(cmd))
        try:
            #print(' '.join(cmd))
            subprocess.check_output(cmd, stderr=subprocess.STDOUT)
            shutil.move(dst, src)
        except subprocess.CalledProcessError as e:
            print('ERROR: this command failed')
            print(' '.join(cmd))
            print(e.output)


def crop_with_gdal_translate(outpath, inpath, ulx, uly, lrx, lry,
                             utm_zone=None, lat_band=None, output_type=None):
    """
    """
    if outpath == inpath:  # hack to allow the output to overwrite the input
        fd, out = tempfile.mkstemp(suffix='.tif', dir=os.path.dirname(inpath))
        os.close(fd)
    else:
        out = outpath

    env = os.environ.copy()

    # these GDAL configuration options speed up the access to remote files
    if inpath.startswith(('http://', 'https://', 's3://', 'gs://')):
        env['CPL_VSIL_CURL_ALLOWED_EXTENSIONS'] = inpath[-3:]
        env['GDAL_DISABLE_READDIR_ON_OPEN'] = 'TRUE'
        env['VSI_CACHE'] = 'TRUE'
        env['GDAL_HTTP_MAX_RETRY'] = '10000'  # needed for storage.googleapis.com 503
        env['GDAL_HTTP_RETRY_DELAY'] = '1'

    # add the relevant "/vsi" prefix to the input url
    if inpath.startswith(('http://', 'https://')):
        path = '/vsicurl/{}'.format(inpath)
    elif inpath.startswith('s3://'):
        env['AWS_REQUEST_PAYER'] = 'requester'
        path = '/vsis3/{}'.format(inpath[len('s3://'):])
    elif inpath.startswith('gs://'):
        path = '/vsicurl/http://storage.googleapis.com/{}'.format(inpath[len('gs://'):])
    else:
        path = inpath

    # build the gdal_translate shell command
    cmd = ['gdal_translate', path, out, '-of', 'GTiff', '-co', 'COMPRESS=DEFLATE',
           '-projwin', str(ulx), str(uly), str(lrx), str(lry)]
    if output_type is not None:
        cmd += ['-ot', output_type]
    if utm_zone is not None:
        if gdal_translate_version() < '2.0':
            print('WARNING: utils.crop_with_gdal_translate argument utm_zone requires gdal >= 2.0')
        else:
            srs = '+proj=utm +zone={}'.format(utm_zone)
            # latitude bands in the southern hemisphere range from 'C' to 'M'
            if lat_band and lat_band < 'N':
                srs += ' +south'
            cmd += ['-projwin_srs', srs]
            #print(' '.join(cmd))

    # run the gdal_translate command
    try:
        subprocess.check_output(cmd, stderr=subprocess.STDOUT, env=env)
    except subprocess.CalledProcessError as e:
        if inpath.startswith(('http://', 'https://')):
            if not requests.head(inpath).ok:
                print('{} is not available'.format(inpath))
                return
        print('ERROR: this command failed')
        print(' '.join(cmd))
        print(e.output)
        return

    if outpath == inpath:  # hack to allow the output to overwrite the input
        shutil.move(out, outpath)


def get_crop_from_aoi(output_path, aoi, metadata_dict, band):
    """
    Crop and download an AOI from a georeferenced image.

    Args:
        output_path (string): path to the output GeoTIFF file
        aoi (geojson.Polygon): area of interest defined by a polygon in longitude, latitude coordinates
        metadata_dict (dict): metadata dictionary
        band (str): desired band, e.g. 'B04' for Sentinel-2 or 'B8' for Landsat-8
    """
    try:  # Sentinel-2
        inpath = metadata_dict['urls']['gcloud'][band]
    except KeyError:  # Landsat-8
        inpath = metadata_dict['assets'][band]['href']
    utm_zone = int(metadata_dict['utm_zone']) if 'utm_zone' in metadata_dict else None
    ulx, uly, lrx, lry, utm_zone, lat_band = utm_bbx(aoi, utm_zone=utm_zone,
                                                     r=60)
    crop_with_gdal_translate(output_path, inpath, ulx, uly, lrx, lry, utm_zone,
                             lat_band)


def crop_with_gdalwarp(outpath, inpath, geojson_path):
    """
    """
    cmd = ['gdalwarp', inpath, outpath, '-ot', 'UInt16', '-of', 'GTiff',
           '-overwrite', '-crop_to_cutline', '-cutline', geojson_path]
    subprocess.check_output(cmd, stderr=subprocess.STDOUT)


def geojson_lonlat_to_utm(aoi):
    """
    """
    # compute the utm zone number of the first polygon vertex
    lon, lat = aoi['coordinates'][0][0]
    utm_zone = utm.from_latlon(lat, lon)[2]

    # convert all polygon vertices coordinates from (lon, lat) to utm
    c = []
    for lon, lat in aoi['coordinates'][0]:
        c.append(utm.from_latlon(lat, lon, force_zone_number=utm_zone)[:2])

    return geojson.Polygon([c])


def utm_bbx(aoi, utm_zone=None, r=None):
    """
    """
    lon, lat = aoi['coordinates'][0][0]
    if utm_zone is None:  # compute the utm zone number of the first vertex
        utm_zone, lat_band = utm.from_latlon(lat, lon)[2:]
    else:
        lat_band = utm.from_latlon(lat, lon, force_zone_number=utm_zone)[3]

    # convert all polygon vertices coordinates from (lon, lat) to utm
    c = []
    for lon, lat in aoi['coordinates'][0]:
        hemisphere = 'north' if lat>=0 else 'south'
        x,y = pyproj.transform(pyproj.Proj('+proj=latlong'), pyproj.Proj('+proj=utm +zone={}{} +{}'.format(utm_zone,lat_band,hemisphere)), lon, lat)
        c.append((x,y))

    # utm bounding box
    bbx = shapely.geometry.Polygon(c).bounds  # minx, miny, maxx, maxy
    ulx, uly, lrx, lry = bbx[0], bbx[3], bbx[2], bbx[1]  # minx, maxy, maxx, miny

    if r is not None:  # round to multiples of the given resolution
        ulx = r * np.round(ulx / r)
        uly = r * np.round(uly / r)
        lrx = r * np.round(lrx / r)
        lry = r * np.round(lry / r)

    return ulx, uly, lrx, lry, utm_zone, lat_band



def latlon_to_pix(path, lat, lon):
   """
   Get the pixel coordinates of a geographic location in a GeoTIFF image.

   Args:
       path: path to the input image
       lat, lon: geographic coordinates of the input location

   Returns:
       x, y: pixel coordinates
   """
   with rasterio.open(path, 'r') as f:
       crs = f.crs
       transform = f.transform

   # transform (lon, lat) to the coordinate reference system of the image
   x, y = pyproj.transform(pyproj.Proj('+proj=latlong'), pyproj.Proj(crs),
                           lon, lat)

   # transform x, y to pixel coordinates
   return ~transform * (x, y)


def latlon_rectangle_centered_at(lat, lon, w, h):
    """
    """
    x, y, number, letter = utm.from_latlon(lat, lon)
    rectangle = [utm.to_latlon(x - .5*w, y - .5*h, number, letter),
                 utm.to_latlon(x - .5*w, y + .5*h, number, letter),
                 utm.to_latlon(x + .5*w, y + .5*h, number, letter),
                 utm.to_latlon(x + .5*w, y - .5*h, number, letter)]
    rectangle.append(rectangle[0])  # close the polygon
    return rectangle


def lonlat_rectangle_centered_at(lon, lat, w, h):
    """
    """
    return [p[::-1] for p in latlon_rectangle_centered_at(lat, lon, w, h)]


def print_elapsed_time(since_first_call=False):
    """
    Print the elapsed time since the last call or since the first call.

    Args:
        since_first_call:
    """
    t2 = datetime.datetime.now()
    if since_first_call:
        print("Total elapsed time:", t2 - print_elapsed_time.t0)
    else:
        try:
            print("Elapsed time:", t2 - print_elapsed_time.t1)
        except AttributeError:
            print("Elapsed time:", t2 - print_elapsed_time.t0)
    print_elapsed_time.t1 = t2
    print()


def show(img):
    """
    """
    import matplotlib.pyplot as plt
    fig, ax = plt.subplots()
    ax.imshow(img, interpolation='nearest')

    def format_coord(x, y):
        col = int(x + 0.5)
        row = int(y + 0.5)
        if col >= 0 and col < img.shape[1] and row >= 0 and row < img.shape[0]:
            z = img[row, col]
            return 'x={}, y={}, z={}'.format(col, row, z)
        else:
            return 'x={}, y={}'.format(col, row)

    ax.format_coord = format_coord
    plt.show()


def warn_with_traceback(message, category, filename, lineno, file=None,
                        line=None):
    traceback.print_stack()
    log = file if hasattr(file,'write') else sys.stderr
    log.write(warnings.formatwarning(message, category, filename, lineno, line))

#warnings.showwarning = warn_with_traceback


def bounding_box2D(pts):
    """
    bounding box for the points pts
    """
    dim = len(pts[0])  # should be 2
    bb_min = [min([t[i] for t in pts]) for i in range(dim)]
    bb_max = [max([t[i] for t in pts]) for i in range(dim)]
    return bb_min[0], bb_min[1], bb_max[0] - bb_min[0], bb_max[1] - bb_min[1]


def rpc_from_geotiff(geotiff_path):
    """
    """
    with rasterio.open(geotiff_path, 'r') as src:
        rpc_dict = src.tags(ns='RPC')
    return rpc_model.RPCModel(rpc_dict)


def points_apply_homography(H, pts):
    """
    Applies an homography to a list of 2D points.

    Args:
        H: numpy array containing the 3x3 homography matrix
        pts: numpy array containing the list of 2D points, one per line

    Returns:
        a numpy array containing the list of transformed points, one per line
    """
    pts = np.asarray(pts)

    # convert the input points to homogeneous coordinates
    if len(pts[0]) < 2:
        print("""points_apply_homography: ERROR the input must be a numpy array
          of 2D points, one point per line""")
        return
    pts = np.hstack((pts[:, 0:2], pts[:, 0:1]*0+1))

    # apply the transformation
    Hpts = (np.dot(H, pts.T)).T

    # normalize the homogeneous result and trim the extra dimension
    Hpts = Hpts * (1.0 / np.tile( Hpts[:, 2], (3, 1)) ).T
    return Hpts[:, 0:2]


def bounding_box_of_projected_aoi(rpc, aoi, z=0, homography=None):
    """
    Return the x, y, w, h pixel bounding box of a projected AOI.

    Args:
        rpc (rpc_model.RPCModel): RPC camera model
        aoi (geojson.Polygon): GeoJSON polygon representing the AOI
        z (float): altitude of the AOI with respect to the WGS84 ellipsoid
        homography (2D array, optional): matrix of shape (3, 3) representing an
            homography to be applied to the projected points before computing
            their bounding box.

    Return:
        x, y (ints): pixel coordinates of the top-left corner of the bounding box
        w, h (ints): pixel dimensions of the bounding box
    """
    lons, lats = np.array(aoi['coordinates'][0]).T
    x, y = rpc.projection(lons, lats, z)
    pts = list(zip(x, y))
    if homography is not None:
        pts = points_apply_homography(homography, pts)
    return np.round(bounding_box2D(pts)).astype(int)


class CropOutside(Exception):
    """
    Exception to raise when attempting to crop outside of the input image.
    """
    pass


def rasterio_crop(filename, x, y, w, h, boundless=True, fill_value=0):
    """
    Read a crop from a file with rasterio and return it as an array.

    This is a working alternative to this rasterio oneliner which currently fails:
    src.read(window=((y, y + h), (x, x + w)), boundless=True, fill_value=0)

    Args:
        filename: path to the input image file
        x, y: pixel coordinates of the top-left corner of the crop
        w, h: width and height of the crop, in pixels
        boundless (bool): similar to gdal_translate "epo: error when partially
            outside" flag. If False, we'll raise an exception when the
            requested crop is not entirely contained within the input image
            bounds. If True, the crop is padded with fill_value.
        fill_value (scalar): constant value used to fill pixels outside of the
            input image.
    """
    with rasterio.open(filename, 'r') as src:
        if not boundless:
            if y < 0 or y + h > src.shape[0] or x < 0 or x + w > src.shape[1]:
                raise CropOutside(('crop {} {} {} {} falls outside of input image '
                                   'whose shape is {}'.format(x, y, w, h, src.shape)))

        crop = fill_value * np.ones((src.count, h, w), dtype=src.profile['dtype'])
        y0 = max(y, 0)
        y1 = min(y + h, src.shape[0])
        x0 = max(x, 0)
        x1 = min(x + w, src.shape[1])
        crop[:, y0 - y:y1 - y, x0 - x:x1 - x] = src.read(window=((y0, y1), (x0, x1)))

    # interleave channels
    return np.moveaxis(crop, 0, 2).squeeze()


def crop_aoi(geotiff, aoi, z=0):
    """
    Crop a geographic AOI in a georeferenced image using its RPC functions.

    Args:
        geotiff (string): path or url to the input GeoTIFF image file
        aoi (geojson.Polygon): GeoJSON polygon representing the AOI
        z (float, optional): base altitude with respect to WGS84 ellipsoid (0
            by default)

    Return:
        crop (array): numpy array containing the cropped image
        x, y, w, h (ints): image coordinates of the crop. x, y are the
            coordinates of the top-left corner, while w, h are the dimensions
            of the crop.
    """
    x, y, w, h = bounding_box_of_projected_aoi(rpc_from_geotiff(geotiff), aoi, z)
    return rasterio_crop(geotiff, x, y, w, h), x, y


def rio_dtype(numpy_dtype):
    """
    Convert a numpy datatype to a rasterio datatype.
    """
    if numpy_dtype == 'bool':
        return rasterio.dtypes.bool_
    elif numpy_dtype == 'uint8':
        return rasterio.dtypes.uint8
    elif numpy_dtype == 'uint16':
        return rasterio.dtypes.uint16
    elif numpy_dtype == 'int16':
        return rasterio.dtypes.int16
    elif numpy_dtype == 'uint32':
        return rasterio.dtypes.uint32
    elif numpy_dtype == 'int32':
        return rasterio.dtypes.int32
    elif numpy_dtype == 'float32':
        return rasterio.dtypes.float32
    elif numpy_dtype == 'float64':
        return rasterio.dtypes.float64
    elif numpy_dtype == 'complex':
        return rasterio.dtypes.complex_
    elif numpy_dtype == 'complex64':
        return rasterio.dtypes.complex64
    elif numpy_dtype == 'complex128':
        return rasterio.dtypes.complex128


def rio_write(path, array, profile={}, tags={}, namespace_tags={}):
    """
    Write a numpy array in a tiff/png/jpeg file with rasterio.

    Args:
        path: path to the output tiff/png/jpeg file
        array: 2D or 3D numpy array containing the image to write
        profile: rasterio profile (ie dictionary of metadata)
        tags: dictionary of additional geotiff tags
        namespace_tags: dictionary of dictionaries of additional geotiff tags
            (e.g.  IMAGE_STRUCTURE, RPC, SUBDATASETS)
    """
    # read image size and number of bands
    if array.ndim > 2:
        height, width, nbands = array.shape
    else:
        nbands = 1
        height, width = array.shape

    # determine the driver based on the file extension
    extension = os.path.splitext(path)[1].lower()
    if extension in ['.tif', '.tiff']:
        driver = 'GTiff'
    elif extension in ['.jpg', '.jpeg']:
        driver = 'jpeg'
    elif extension in ['.png']:
        driver = 'png'
    else:
        print('ERROR: unknown extension {}'.format(extension))

    with warnings.catch_warnings():  # noisy may occur here
        warnings.filterwarnings("ignore", category=FutureWarning)
        profile.update(driver=driver, count=nbands, width=width, height=height,
                       dtype=rio_dtype(array.dtype), quality=100)
        with rasterio.open(path, 'w', **profile) as dst:
            if array.ndim > 2:
                dst.write(np.moveaxis(array, 2, 0))
            else:
                dst.write(np.array([array]))
            dst.update_tags(**tags)
            for k, v in namespace_tags.items():
                dst.update_tags(ns=k, **v)



#!/usr/bin/env python3
# vim: set fileencoding=utf-8
# pylint: disable=C0103

"""
Search of Sentinel images.

Copyright (C) 2016-18, Carlo de Franchis <carlo.de-franchis@ens-cachan.fr>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function
import os
import sys
import argparse
import datetime
import json
import shapely.geometry
import shapely.wkt
import requests
import dateutil.parser

from tsd import utils


# http://sentinel-s2-l1c.s3-website.eu-central-1.amazonaws.com
API_URLS = {
    'copernicus': 'https://scihub.copernicus.eu/dhus/',
    'austria': 'https://data.sentinel.zamg.ac.at/',
    'finland': 'https://finhub.nsdc.fmi.fi/'
}


def read_copernicus_credentials_from_environment_variables():
    """
    Read the user Copernicus Open Access Hub credentials.
    """
    try:
        login = os.environ['COPERNICUS_LOGIN']
        password = os.environ['COPERNICUS_PASSWORD']
    except KeyError as e:
        login = 'tsd-default-user'
        password = 'b3c5e714034282ea5c'
    return login, password


def post_scihub(url, query, user, password):
    """
    Send a POST request to scihub.
    """
    r = requests.post(url, dict(q=query), auth=(user, password))
    if r.ok:
        return r
    else:
        print('ERROR:', end=' ')
        if r.status_code == 503:
            print('The Sentinels Scientific Data Hub is down. Check on'
                  ' https://scihub.copernicus.eu/dhus/#/home')
        elif r.status_code == 401:
            print('Authentication failed with', user, password)
        else:
            print('Scientific Data Hub returned error', r.status_code)
        r.raise_for_status()


def build_scihub_query(aoi, start_date=None, end_date=None,
                       satellite='Sentinel-1', product_type='GRD',
                       operational_mode='IW', relative_orbit_number=None,
                       swath_identifier=None):
    """
    """
    # default start/end dates
    if end_date is None:
        end_date = datetime.datetime.now()
    if start_date is None:
        start_date = datetime.datetime(2000, 1, 1)

    # build the url used to query the scihub API
    query = 'platformname:{}'.format(satellite)
    query += ' AND producttype:{}'.format(product_type)
    if satellite == 'Sentinel-1':
        query += ' AND sensoroperationalmode:{}'.format(operational_mode)
    query += ' AND beginposition:[{}Z TO {}Z]'.format(start_date.isoformat(),
                                                      end_date.isoformat())
    if relative_orbit_number is not None:
        query += ' AND relativeorbitnumber:{}'.format(relative_orbit_number)

    if swath_identifier is not None:
        query += ' AND swathidentifier:{}'.format(swath_identifier)

    # queried polygon or point
    # http://forum.step.esa.int/t/advanced-search-in-data-hub-contains-intersects/1150/2
    query += ' AND footprint:\"contains({})\"'.format(shapely.geometry.shape(aoi).wkt)

    return query


def load_query(query, api_url, start_row=0, page_size=100):
    """
    Do a full-text query on the SciHub API using the OpenSearch format.

    https://scihub.copernicus.eu/twiki/do/view/SciHubUserGuide/3FullTextSearch
    """
    # load query results
    url = '{}search?format=json&rows={}&start={}'.format(api_url, page_size,
                                                         start_row)
    login, password = read_copernicus_credentials_from_environment_variables()
    r = post_scihub(url, query, login, password)

    # parse response content
    d = r.json()['feed']
    total_results = int(d['opensearch:totalResults'])
    entries = d.get('entry', [])

    # if the query returns only one product entries will be a dict, not a list
    if isinstance(entries, dict):
        entries = [entries]

    # repeat query until all results have been loaded
    output = entries
    if total_results > start_row + page_size:
        output += load_query(query, api_url, start_row=(start_row + page_size))
    return output


def prettify_scihub_dict(d):
    """
    Convert the oddly formatted json response of scihub into something nicer.

    Scihub json response is roughly a dict with a key per datatype (int, str,
    date, ...). The value associated to each key is a list of dicts like
    {'name': foo, 'content': bar}, while it would be more natural to have
    (foo, bar) as a (key, value) pair of the main json dict.

    Args:
        d (dict): json-formatted metadata returned by scihub for a single SAFE

    Returns:
        prettified metadata dict
    """
    out = {}
    for k in ['title', 'id', 'summary']:
        if k in d:
            out[k] = d[k]

    if 'int' in d:
        for x in d['int']:
            out[x['name']] = int(x['content'])

    if 'str' in d:
        for x in d['str']:
            out[x['name']] = x['content']

    if 'date' in d:
        for x in d['date']:
            out[x['name']] = x['content']

    if 'link' in d:
        out['links'] = {}
        for x in d['link']:
            if 'rel' in x:
                out['links'][x['rel']] = x['href']
            else:
                out['links']['main'] = x['href']
    return out


def search(aoi, start_date=None, end_date=None, satellite='Sentinel-1',
           product_type='GRD', operational_mode='IW',
           relative_orbit_number=None, swath_identifier=None,
           api='copernicus'):
    """
    List the Sentinel images covering a location using Copernicus Scihub API.
    """
    if satellite == 'Sentinel-2' and product_type not in ['S2MSI1C', 'S2MSI2A', 'S2MSI2Ap']:
        product_type = 'S2MSI1C'

    query = build_scihub_query(aoi, start_date, end_date, satellite,
                               product_type, operational_mode,
                               relative_orbit_number, swath_identifier)
    results = [prettify_scihub_dict(x) for x in load_query(query, API_URLS[api])]

    # check if the image footprint contains the area of interest
    not_covering = []
    aoi_shape = shapely.geometry.shape(aoi)
    for x in results:
        if not shapely.wkt.loads(x['footprint']).contains(aoi_shape):
            not_covering.append(x)

    for x in not_covering:
        results.remove(x)

    return results


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description=('Search of Sentinel images'))
    parser.add_argument('--geom', type=utils.valid_geojson,
                        help=('path to geojson file'))
    parser.add_argument('--lat', type=utils.valid_lat,
                        help=('latitude of the center of the rectangle AOI'))
    parser.add_argument('--lon', type=utils.valid_lon,
                        help=('longitude of the center of the rectangle AOI'))
    parser.add_argument('-w', '--width', type=int, default=5000,
                        help='width of the AOI (m), default 5000 m')
    parser.add_argument('-l', '--height', type=int, default=5000,
                        help='height of the AOI (m), default 5000 m')
    parser.add_argument('-s', '--start-date', type=utils.valid_datetime,
                        help='start date, YYYY-MM-DD')
    parser.add_argument('-e', '--end-date', type=utils.valid_datetime,
                        help='end date, YYYY-MM-DD')
    parser.add_argument('--satellite', default='Sentinel-1',
                        help='which satellite: Sentinel-1 or Sentinel-2')
    parser.add_argument('--product-type', default='GRD',
                        help='type of image: RAW, SLC, GRD, OCN (for S1), S2MSI1C, S2MSI2A, S2MSI2Ap (for S2)')
    parser.add_argument('--operational-mode', default='IW',
                        help='(for S1) acquisiton mode: SM, IW, EW or WV')
    parser.add_argument('--swath-identifier',
                        help='(for S1) subswath id: S1..S6 or IW1..IW3 or EW1..EW5')
    parser.add_argument('--api', default='copernicus',
                        help='mirror to use: copernicus, austria or finland')
    args = parser.parse_args()

    if args.geom and (args.lat or args.lon):
        parser.error('--geom and {--lat, --lon} are mutually exclusive')

    if not args.geom and (not args.lat or not args.lon):
        parser.error('either --geom or {--lat, --lon} must be defined')

    if args.geom:
        aoi = args.geom
    else:
        aoi = utils.geojson_geometry_object(args.lat, args.lon, args.width,
                                            args.height)

    print(json.dumps(search(aoi, args.start_date, args.end_date,
                            satellite=args.satellite,
                            product_type=args.product_type,
                            operational_mode=args.operational_mode,
                            swath_identifier=args.swath_identifier,
                            api=args.api)))



#!/usr/bin/env python3
# vim: set fileencoding=utf-8
# pylint: disable=C0103

"""
Search of Planet images.

Copyright (C) 2016-18, Carlo de Franchis <carlo.de-franchis@m4x.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function
import os
import argparse
import datetime
import json
import sys
import shapely.geometry
import dateutil.parser
from planet import api

from tsd import utils

client = api.ClientV1()
ITEM_TYPES = ['PSScene3Band', 'PSScene4Band', 'PSOrthoTile', 'REScene', 'REOrthoTile',
              'Sentinel2L1C', 'Landsat8L1G', 'Sentinel1', 'SkySatScene']


def search(aoi, start_date=None, end_date=None, item_types=ITEM_TYPES,
           satellite_id=None, search_type='contains', remove_duplicates=True):
    """
    Search for images using Planet API.

    Args:
        aoi: geojson.Polygon or geojson.Point object
        item_types: list of strings.
        satellite_id (str): satellite identifier, e.g. '0f02'
        search_type (str): either 'intersects' or 'contains'

    """
    # default start/end dates
    if end_date is None:
        end_date = datetime.datetime.now()
    if start_date is None:
        start_date = end_date - datetime.timedelta(365)

    # build a search query with filters for the AOI and the date range
    geom_filter = api.filters.geom_filter(aoi)
    date_filter = api.filters.date_range('acquired', gte=start_date, lte=end_date)
    if 'PSScene3Band' in item_types or 'PSScene4Band' in item_types:
        quality_filter = api.filters.string_filter('quality_category', 'standard')
        query = api.filters.and_filter(geom_filter, date_filter, quality_filter)
    else:
        query = api.filters.and_filter(geom_filter, date_filter)

    if satellite_id:
        query = api.filters.and_filter(query,
                                       api.filters.string_filter('satellite_id',
                                                                 satellite_id))

    request = api.filters.build_search_request(query, item_types)

    # this will cause an exception if there are any API related errors
    try:
        response = client.quick_search(request)
    except api.exceptions.InvalidAPIKey as e:
        print("\nERROR: The {} module requires".format(os.path.basename(__file__)),
              "the PL_API_KEY environment variable to be defined with valid",
              "credentials for https://www.planet.com/. Create an account if",
              "you don't have one (it's free) then edit the relevant configuration",
              "files (eg .bashrc) to define this environment variable.\n")
        raise e

    # list results
    aoi = shapely.geometry.shape(aoi)
    results = []
    for x in response.items_iter(limit=None):
        if search_type == 'contains':  # keep only images containing the full AOI
            if not shapely.geometry.shape(x['geometry']).contains(aoi):
                continue
        results.append(x)

    # sort results by acquisition date
    dates = [dateutil.parser.parse(x['properties']['acquired']) for x in results]
    results = [r for d, r in sorted(zip(dates, results), key=lambda t:t[0])]
    dates.sort()

    # remove duplicates (two images are said to be duplicates if within 5 minutes)
    if remove_duplicates:
        to_remove = []
        for i, (d, r) in enumerate(list(zip(dates, results))[:-1]):
            if dates[i+1] - d < datetime.timedelta(seconds=300):
                to_remove.append(r)

    return [r for r in results if r not in to_remove]


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Search of images through Planet API.')
    parser.add_argument('--geom', type=utils.valid_geojson,
                        help=('path to geojson file'))
    parser.add_argument('--lat', type=utils.valid_lat,
                        help=('latitude of the center of the rectangle AOI'))
    parser.add_argument('--lon', type=utils.valid_lon,
                        help=('longitude of the center of the rectangle AOI'))
    parser.add_argument('-w', '--width', type=int, default=5000,
                        help='width of the AOI (m), default 5000 m')
    parser.add_argument('-l', '--height', type=int, default=5000,
                        help='height of the AOI (m), default 5000 m')
    parser.add_argument('-s', '--start-date', type=utils.valid_datetime,
                        help='start date, YYYY-MM-DD')
    parser.add_argument('-e', '--end-date', type=utils.valid_datetime,
                        help='end date, YYYY-MM-DD')
    parser.add_argument('--search-type', choices=['contains', 'intersects'],
                        default='contains', help='search type')
    parser.add_argument('--satellite-id', help='satellite identifier, e.g. 0f02')
    parser.add_argument('--keep-duplicates', action='store_true',
                        help='keep all images even when two were acquired within'
                             ' less than 5 minutes (the default behaviour is to'
                             ' discard such duplicates)')
    parser.add_argument('--item-types', nargs='*', choices=ITEM_TYPES,
                        default=['PSScene3Band'], metavar='',
                        help=('space separated list of item types to'
                              ' search for. Default is PSScene3Band. Allowed'
                              ' values are {}'.format(', '.join(ITEM_TYPES))))
    args = parser.parse_args()

    if args.geom and (args.lat or args.lon):
        parser.error('--geom and {--lat, --lon} are mutually exclusive')

    if not args.geom and (not args.lat or not args.lon):
        parser.error('either --geom or {--lat, --lon} must be defined')

    if args.geom:
        aoi = args.geom
    else:
        aoi = utils.geojson_geometry_object(args.lat, args.lon, args.width,
                                            args.height)

    print(json.dumps(search(aoi, start_date=args.start_date,
                            end_date=args.end_date,
                            item_types=args.item_types,
                            search_type=args.search_type,
                            satellite_id=args.satellite_id,
                            remove_duplicates=~args.keep_duplicates)))



import os
import argparse
import datetime
import json
import requests
import shapely.geometry
import numpy as np
import pandas as pd

# from pandas.io import gbq
from google.cloud import bigquery

from bs4 import BeautifulSoup
from tqdm import tqdm
import utils

from pyproj import Proj, transform
from json.decoder import JSONDecodeError


def parse_url(url):
    _, file = url.split('gs://')
    bucket, *prefix = file.split('/')
    return bucket, '/'.join(prefix), prefix[-1]


def get_footprint(img, source='roda'):
    mgrs = img['mgrs_tile']
    date = pd.to_datetime(img['sensing_time'])

    if source=='roda':
        url = 'https://roda.sentinel-hub.com/sentinel-s2-l1c/tiles/{}/{}/{}/{}/{}/{}/0/tileInfo.json'.format(mgrs[:2].lstrip('0'),
                                                                                                             mgrs[2],
                                                                                                             mgrs[3:],
                                                                                                             date.year,
                                                                                                             date.month,
                                                                                                             date.day)
        try:
            metadata = requests.get(url).json()
        except JSONDecodeError:
            return get_footprint(img, source='google')

        key = 'tileDataGeometry' if 'tileDataGeometry' in metadata else 'tileGeometry'
        epsg = metadata[key]['crs']['properties']['name'].split(':')[-1]
        utm_coords = metadata[key]['coordinates'][0]

    else:
        # Source should be google
        if '.' in img['granule_id']:
            # Old format
            filename = '{}.xml'.format(img['product_id'].replace('PRD_MSIL1C', 'MTD_SAFL1C'))
        else:
            # New format
            filename = 'MTD_MSIL1C.xml'

        url = '{}/{}'.format(img['base_url'].replace('gs://', 'http://storage.googleapis.com/'), filename)
        r = requests.get(url)
        coords = BeautifulSoup(r.content, 'lxml').find('ext_pos_list').text.strip().split(' ')
        coords = [(float(coords[2*i]),float(coords[2*i+1])) for i in range(int(len(coords)/2))]
        ref_lat, ref_lon = coords[0]
        _,_,zn,_ = utm.from_latlon(ref_lat, ref_lon)
        utm_coords = [utm.from_latlon(x,y)[:2] for x,y in coords]
        epsg = str(32700-round((45+ref_lat)/90)*100+zn)

    return shapely.geometry.Polygon(utm_coords), epsg


def convert_aoi_to_utm(aoi, epsg):
    outProj = Proj(init='epsg:{}'.format(epsg))
    inProj = Proj(init='epsg:4326')
    utm_aoi = []
    for x, y in aoi['coordinates'][0]:
        utm_aoi.append(transform(inProj, outProj, x, y))
    return shapely.geometry.Polygon(utm_aoi)


def query_string(lat, lon, start_date, end_date, satellite, sensor):
    if end_date is None:
        end_date = datetime.date.today()
    if start_date is None:
        start_date = end_date - datetime.timedelta(365)

    date_query = 'sensing_time >= "{}" AND sensing_time <= "{}"'.format(start_date, end_date)
    loc_query = 'north_lat>={} AND south_lat<={} AND west_lon<={} AND east_lon>={}'.format(lat, lat, lon, lon)
    additional_query = ''

    if satellite=='Sentinel-2':
        tab_name = '`bigquery-public-data.cloud_storage_geo_index.sentinel_2_index`'
    elif 'Landsat' in satellite:
        tab_name = '`bigquery-public-data.cloud_storage_geo_index.landsat_index`'
        if sensor is not None:
            sensor_query = ' AND sensor_id="{}"'.format(sensor.replace('OLITIRS', 'OLI_TIRS'))
            additional_query += sensor_query
        if '-' in satellite:
            # Specific query for one Landsat
            sat_query = ' AND spacecraft_id="{}"'.format(satellite.upper().replace('-', '_'))
            additional_query += sat_query
    else:
        raise KeyError('Wrong Satellite name, you entered {}'.format(satellite))

    query = 'SELECT * FROM {} WHERE {} AND {}{}'.format(tab_name, date_query, loc_query, additional_query)

    return query


def search(aoi, start_date=None, end_date=None, satellite='Sentinel-2', sensor=None):
    """
    List images covering an area of interest (AOI) using Google Index.

    Args:
        aoi: geojson.Polygon or geojson.Point object
    """
    # compute the centroid of the area of interest
    lon, lat = shapely.geometry.shape(aoi).centroid.coords.xy
    lon, lat = lon[0], lat[0]

    # build query
    query = query_string(lat, lon, start_date, end_date, satellite, sensor)

    # query Gcloud BigQuery Index
    try:
        private_key = os.environ['GOOGLE_APPLICATION_CREDENTIALS']
    except KeyError as e:
        print('You must have the env variable GOOGLE_APPLICATION_CREDENTIALS linking to the cred json file')
        raise e

    # df = gbq.read_gbq(query, private_key=private_key)
    client = bigquery.Client.from_service_account_json(private_key)
    rows = list(client.query(query).result())
    df = pd.DataFrame(dict(row.items()) for row in rows)

    # check if the image footprint contains the area of interest
    if satellite=='Sentinel-2':
        res = []
        for i, row in df.iterrows():
            footprint, epsg = get_footprint(row)
            utm_aoi = convert_aoi_to_utm(aoi, epsg)
            if footprint.contains(utm_aoi):
                res.append(row.to_dict())
    else:
        # We need to remove duplicates
        order_collection_category = {'T1':0, 'T2':1, 'T3':2, 'RT':3, 'N/A':4}
        order_collection_number = {'01':0, 'PRE':1}
        df['order_collection_category'] = df['collection_category'].apply(lambda x: order_collection_category[x])
        df['order_collection_number'] = df['collection_number'].apply(lambda x: order_collection_number[x])
        unique_scene = ['wrs_path', 'wrs_row', 'spacecraft_id', 'sensor_id', 'date_acquired']
        orders = ['order_collection_number', 'order_collection_category']
        df.sort_values(by=unique_scene+orders, inplace=True)
        res = df.groupby(unique_scene).first().reset_index().drop(orders, axis=1).sort_values(by=['date_acquired']).to_dict('records')
    return res


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Search of Sentinel-2 images.')
    parser.add_argument('--satellite', choices=['Sentinel-2', 'Landsat', 'Landsat-4', 'Landsat-5', 'Landsat-6', 'Landsat-7', 'Landsat-8'],
                        help=('either all "Landsat", one specific "Landsat-8" or "Sentinel-2"'),
                        default='Sentinel-2')
    parser.add_argument('--sensor', choices=['MSS', 'TM', 'ETM', 'OLITIRS'],
                        help=('Only for Landsat, see https://landsat.usgs.gov/what-are-band-designations-landsat-satellites'),
                        default=None)
    parser.add_argument('--geom', type=utils.valid_geojson,
                        help=('path to geojson file'))
    parser.add_argument('--lat', type=utils.valid_lat,
                        help=('latitude of the center of the rectangle AOI'))
    parser.add_argument('--lon', type=utils.valid_lon,
                        help=('longitude of the center of the rectangle AOI'))
    parser.add_argument('-w', '--width', type=int, default=5000,
                        help='width of the AOI (m), default 5000 m')
    parser.add_argument('-l', '--height', type=int, default=5000,
                        help='height of the AOI (m), default 5000 m')
    parser.add_argument('-s', '--start-date', type=utils.valid_datetime,
                        help='start date, YYYY-MM-DD')
    parser.add_argument('-e', '--end-date', type=utils.valid_datetime,
                        help='end date, YYYY-MM-DD')
    args = parser.parse_args()

    if args.geom and (args.lat or args.lon):
        parser.error('--geom and {--lat, --lon} are mutually exclusive')

    if not args.geom and (not args.lat or not args.lon):
        parser.error('either --geom or {--lat, --lon} must be defined')

    if args.geom:
        aoi = args.geom
    else:
        aoi = utils.geojson_geometry_object(args.lat, args.lon, args.width,
                                            args.height)

    print(json.dumps(search(aoi, start_date=args.start_date,
                            end_date=args.end_date, satellite=args.satellite,
                            sensor=args.sensor)))



import os
import sys
import geojson
import kml2geojson


def keep_only_polygons_from_geometry_collection(g):
    """
    Remove all elements that aren't Polygons from the 'geometries' list.
    """
    g['geometries'] = [x for x in g['geometries'] if x['type'] == 'Polygon']


def remove_z_from_polygon_coordinates(p):
    """
    Remove the third (z) coordinate from the points of a polygon.
    """
    p['coordinates'] = [[x[:2] for x in p['coordinates'][0]]]


def main(kml_filename, verbose=False):
    """
    Extract information from the kml file distributed by ESA to describe the
    Sentinel-2 MGRS tiling grid.

    This file is distributed on ESA Sentinel website at:

    https://sentinel.esa.int/documents/247904/1955685/S2A_OPER_GIP_TILPAR_MPC__20151209T095117_V20150622T000000_21000101T000000_B00.kml
    """
    kml2geojson.main.convert(kml_filename, 's2_mgrs_grid')
    with open(os.path.join('s2_mgrs_grid', kml_filename.replace('.kml', '.geojson')), 'r') as f:
        grid = geojson.load(f)

    mgrs_tiles = []
    for x in grid['features']:
        g = x['geometry']
        keep_only_polygons_from_geometry_collection(g)
        for p in g['geometries']:
            remove_z_from_polygon_coordinates(p)
        mgrs_id = x['properties']['name']
        mgrs_tiles.append(geojson.Feature(id=mgrs_id, geometry=g))
        if verbose:
            print(mgrs_id, end=' ')
            print(g)

    return geojson.FeatureCollection(mgrs_tiles)


if __name__ == '__main__':
    if len(sys.argv) != 2:
        print("\t Usage: {} input.kml > output.geojson".format(sys.argv[0]))
    else:
        print(geojson.dumps(main(sys.argv[1]), indent=2))



#!/usr/bin/env python3
# vim: set fileencoding=utf-8
# pylint: disable=C0103

"""
Automatic crop and download of Landsat images.

Copyright (C) 2016-18, Carlo de Franchis <carlo.de-franchis@m4x.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import re
import os
import sys
import shutil
import argparse
import multiprocessing
import dateutil.parser
import datetime
import requests
import bs4
import boto3
import botocore
import geojson
import shapely.geometry
import rasterio
import numpy as np

from tsd import utils
from tsd import parallel
from tsd import l8_metadata_parser

# list of spectral bands
ALL_BANDS = ['B{}'.format(i) for i in range(1,12)] + ['BQA']

def is_image_empty(path, bands):
    band_files = [path.format(band) for band in bands if band!='BQA']
    for band_file in band_files:
        nonzero = rasterio.open(band_file).read().sum()
        # print('There are {} pixels in {}'.format(nonzero, band_file))
        if nonzero==0:
            return True
        else:
            continue
    return False


def check_args(api, mirror):
    if mirror == 'gcloud' and api not in ['gcloud', 'devseed']:
        raise ValueError("ERROR: You must use gcloud or devseed api to use gcloud as mirror")
    if api == 'gcloud':
        try:
            private_key = os.environ['GOOGLE_APPLICATION_CREDENTIALS']
        except KeyError:
            raise ValueError('You must have the env variable GOOGLE_APPLICATION_CREDENTIALS linking to the cred json file')


def search(aoi, start_date=None, end_date=None, satellite='L8',
           sensor='OLITIRS', api='devseed'):
    """
    Search Landsat images covering an AOI and timespan using a given API.

    Args:
        aoi (geojson.Polygon): area of interest
        start_date (datetime.datetime): start of the search time range
        end_date (datetime.datetime): end of the search time range
        api (str, optional): either gcloud (default), scihub, planet or devseed
        satellite (str, optional): either L1...L8
        sensor (str, optional): MSS, TM, ETM, OLITIRS
        see https://landsat.usgs.gov/what-are-band-designations-landsat-satellites

    Returns:
        list of image objects
    """
    # list available images
    if api == 'gcloud':
        from tsd import search_gcloud
        images = search_gcloud.search(aoi, start_date, end_date, satellite=satellite, sensor=sensor)
        images = [l8_metadata_parser.LandsatGcloudParser(img) for img in images]
    elif api == 'devseed':
        from tsd import search_devseed
        images = search_devseed.search(aoi, start_date, end_date, satellite='Landsat-8')
        images = [l8_metadata_parser.LandsatDevSeedParser(img) for img in images]

    # sort images by acquisition day, then by mgrs id
    images.sort(key=(lambda k: (k.date.date(), k.row, k.path)))

    print('Found {} images'.format(len(images)))
    return images


def download(imgs, bands, aoi, mirror, out_dir, parallel_downloads):
    """
    Download a timeseries of crops with GDAL VSI feature.

    Args:
        imgs (list): list of images
        bands (list): list of bands
        aoi (geojson.Polygon): area of interest
        mirror (str): either 'aws' or 'gcloud'
        out_dir (str): path where to store the downloaded crops
        parallel_downloads (int): number of parallel downloads
    """
    coords = utils.utm_bbx(aoi)
    crops_args = []
    for img in imgs:
        for b in set(bands + ['BQA']):
            fname = os.path.join(out_dir, '{}_band_{}.tif'.format(img.filename, b))
            crops_args.append((fname, img.urls[mirror][b]))

    out_dir = os.path.abspath(os.path.expanduser(out_dir))
    os.makedirs(out_dir, exist_ok=True)
    print('Downloading {} crops ({} images with {} bands)...'.format(len(crops_args),
                                                                     len(imgs),
                                                                     len(bands) +1),
          end=' ')
    parallel.run_calls(utils.crop_with_gdal_translate, crops_args,
                       extra_args=(*coords,), pool_type='threads',
                       nb_workers=parallel_downloads)


def bands_files_are_valid(img, bands, api, directory):
    """
    Check if all bands images files are valid.
    """
    filenames = ['{}_band_{}.tif'.format(img.filename, b) for b in bands]
    paths = [os.path.join(directory, f) for f in filenames]
    return all(utils.is_valid(p) for p in paths)


def is_image_cloudy(qa_band_file, p=0.5):
    """
    Tell if a Landsat-8 image crop is cloud-covered according to the QA band.

    The crop is considered covered if a fraction larger than p of its pixels are
    labeled as clouds in the QA band.

    Args:
        qa_band_file: path to a Landsat-8 QA band crop
        p: fraction threshold
    """
    with rasterio.open(qa_band_file, 'r') as f:
        x = f.read(1)
    bqa_cloud_yes = [61440, 59424, 57344, 56320, 53248]
    bqa_cloud_maybe = [39936, 36896, 36864]
    mask = np.in1d(x, bqa_cloud_yes + bqa_cloud_maybe).reshape(x.shape)
    return np.count_nonzero(mask) > p * x.size


def read_cloud_masks(imgs, bands, parallel_downloads, p=0.5,
                     out_dir=''):
    """
    Read Landsat-8 cloud masks and intersects them with the input aoi.

    Args:
        imgs (list): list of images
        bands (list): list of bands
        parallel_downloads (int): number of parallel gml files downloads
        p (float): cloud area threshold above which our aoi is said to be
            'cloudy' in the current image
    """
    print('Reading {} cloud masks...'.format(len(imgs)), end=' ')
    names = [img.filename for img in imgs]
    qa_names = [os.path.join(out_dir, '{}_band_BQA.tif'.format(f)) for f in names]
    cloudy = parallel.run_calls(is_image_cloudy, qa_names,
                                pool_type='processes',
                                nb_workers=parallel_downloads, verbose=False)
    print('{} cloudy images out of {}'.format(sum(cloudy), len(imgs)))

    for img, cloud in zip(imgs, cloudy):
        if cloud:
            os.makedirs(os.path.join(out_dir, 'cloudy'), exist_ok=True)
            for b in list(set(bands + ['BQA'])):
                f = '{}_band_{}.tif'.format(img.filename, b)
                shutil.move(os.path.join(out_dir, f),
                            os.path.join(out_dir, 'cloudy', f))


def read_empty_images(imgs, bands, parallel_downloads,
                     out_dir=''):
    """
    Check whether image is empty, since we do not have access to the footprint.

    Args:
        imgs (list): list of images
        bands (list): list of bands
        parallel_downloads (int): number of parallel gml files downloads
    """
    print('Reading {} QA masks...'.format(len(imgs)), end=' ')
    names = [img.filename for img in imgs]
    base_names = [os.path.join(out_dir, '{}_band_{{}}.tif'.format(f)) for f in names]
    cloudy = parallel.run_calls(is_image_empty, base_names,
                                pool_type='processes',
                                extra_args=(bands,),
                                nb_workers=parallel_downloads, verbose=False)
    print('{} empty images out of {}'.format(sum(cloudy), len(imgs)))

    for img, cloud in zip(imgs, cloudy):
        if cloud is True:
            os.makedirs(os.path.join(out_dir, 'empty'), exist_ok=True)
            for b in list(set(bands + ['BQA'])):
                f = '{}_band_{}.tif'.format(img.filename, b)
                shutil.move(os.path.join(out_dir, f),
                            os.path.join(out_dir, 'empty', f))


def get_time_series(aoi, start_date=None, end_date=None, bands=['B8'],
                    satellite='Landsat', sensor=None,
                    out_dir='', api='devseed', mirror='gcloud',
                    cloud_masks=False, check_empty=False,
                    parallel_downloads=multiprocessing.cpu_count()):
    """
    Main function: crop and download a time series of Sentinel-2 images.

    Args:
        aoi (geojson.Polygon): area of interest
        start_date (datetime.datetime, optional): start of the time range
        end_date (datetime.datetime, optional): end of the time range
        bands (list, optional): list of bands
        satellite (str, optional): either L1...L8
        sensor (str, optional): MSS, TM, ETM, OLI
        see https://landsat.usgs.gov/what-are-band-designations-landsat-satellites
        out_dir (str, optional): path where to store the downloaded crops
        api (str, optional): either devseed (default), scihub, planet or gcloud
        mirror (str, optional): either 'aws' or 'gcloud'
        cloud_masks (bool, optional): if True, cloud masks are downloaded and
            cloudy images are discarded
        check_empty (bool, optional): if True, QA masks are downloaded and
            empty images are discarded
        parallel_downloads (int): number of parallel gml files downloads
    """
    # check access to the selected search api and download mirror
    check_args(api, mirror)

    # default date range
    if end_date is None:
        end_date = datetime.date.today()
    if start_date is None:
        start_date = end_date - datetime.timedelta(91)  # 3 months

    # list available images
    images = search(aoi, start_date, end_date, satellite, sensor, api=api)

    # download crops
    download(images, bands, aoi, mirror, out_dir, parallel_downloads)

    # discard images that failed to download
    images = [i for i in images if bands_files_are_valid(i, bands, api, out_dir)]

    # embed all metadata as GeoTIFF tags in the image files
    for img in images:
        metadata = vars(img)
        metadata['downloaded_by'] = 'TSD on {}'.format(datetime.datetime.now().isoformat())
        for b in bands:
            filepath = os.path.join(out_dir, '{}_band_{}.tif'.format(img.filename, b))
            utils.set_geotif_metadata_items(filepath, metadata)

    if cloud_masks:  # discard images that are totally covered by clouds
        read_cloud_masks(images, bands, parallel_downloads,
                         out_dir=out_dir)

    if check_empty:  # discard images that are totally empty
        read_empty_images(images, bands, parallel_downloads,
                         out_dir=out_dir)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description=('Automatic download and crop '
                                                  'of Landsat images'))
    parser.add_argument('--geom', type=utils.valid_geojson,
                        help=('path to geojson file'))
    parser.add_argument('--lat', type=utils.valid_lat,
                        help=('latitude of the center of the rectangle AOI'))
    parser.add_argument('--lon', type=utils.valid_lon,
                        help=('longitude of the center of the rectangle AOI'))
    parser.add_argument('-w', '--width', type=int, default=5000,
                        help='width of the AOI (m), default 5000 m')
    parser.add_argument('-l', '--height', type=int, default=5000,
                        help='height of the AOI (m), default 5000 m')
    parser.add_argument('-s', '--start-date', type=utils.valid_datetime,
                        help='start date, YYYY-MM-DD')
    parser.add_argument('-e', '--end-date', type=utils.valid_datetime,
                        help='end date, YYYY-MM-DD')
    parser.add_argument('-b', '--band', nargs='*', default=['B8'],
                        choices=ALL_BANDS + ['all'], metavar='',
                        help=('space separated list of spectral bands to'
                              ' download. Default is 8 (panchro). Allowed values'
                              ' are {}'.format(', '.join(ALL_BANDS))))
    parser.add_argument('-o', '--outdir', type=str, help=('path to save the '
                                                          'images'), default='')
    parser.add_argument('--api', type=str, choices=['devseed', 'planet', 'scihub', 'gcloud'],
                        default='devseed', help='search API')
    parser.add_argument('--mirror', type=str, choices=['aws', 'gcloud'],
                        default='gcloud', help='download mirror')
    parser.add_argument('--satellite', type=str, choices=['Landsat', 'Landsat-4', 'Landsat-5', 'Landsat-6', 'Landsat-7', 'Landsat-8'],
                        default='Landsat', help='satellite')
    parser.add_argument('--sensor', type=str, choices=['MSS', 'TM', 'ETM', 'OLITIRS'],
                        default=None, help='sensor')
    parser.add_argument('--parallel-downloads', type=int,
                        default=multiprocessing.cpu_count(),
                        help='max number of parallel crops downloads')
    parser.add_argument('--cloud-masks',  action='store_true',
                        help=('download cloud masks crops from provided GML files'))
    parser.add_argument('--check-empty',  action='store_true',
                        help=('download QA masks to remove empty images'))
    args = parser.parse_args()

    if 'all' in args.band:
        args.band = ALL_BANDS

    if args.geom and (args.lat or args.lon):
        parser.error('--geom and {--lat, --lon} are mutually exclusive')

    if not args.geom and (not args.lat or not args.lon):
        parser.error('either --geom or {--lat, --lon} must be defined')

    if args.geom:
        aoi = args.geom
    else:
        aoi = utils.geojson_geometry_object(args.lat, args.lon, args.width,
                                            args.height)
    get_time_series(aoi, start_date=args.start_date, end_date=args.end_date,
                    bands=args.band,
                    satellite=args.satellite, sensor=args.sensor,
                    out_dir=args.outdir, api=args.api,
                    mirror=args.mirror,
                    cloud_masks=args.cloud_masks,
                    check_empty=args.check_empty,
                    parallel_downloads=args.parallel_downloads)



#!/usr/bin/env python3
# vim: set fileencoding=utf-8
# pylint: disable=C0103

"""
Search of Landsat-8 and Sentinel-2 images using Development Seed API.

Copyright (C) 2016-19, Carlo de Franchis <carlo.de-franchis@m4x.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import argparse
import datetime
import json
import shapely.geometry

import satsearch

from tsd import utils


def search(aoi, start_date=None, end_date=None, satellite='Landsat-8'):
    """
    List images covering an area of interest (AOI) using Development Seed’s API.

    Args:
        aoi: geojson.Polygon or geojson.Point object
        satellite: either Landsat-8 or Sentinel-2
    """
    # date range
    if end_date is None:
        end_date = datetime.date.today()
    if start_date is None:
        start_date = end_date - datetime.timedelta(365)

    # collection
    if satellite.lower() in ['sentinel-2', 'sentinel2', 'sentinel']:
        collection = 'sentinel-2-l1c'
    elif satellite.lower() in ['landsat-8', 'landsat8', 'landsat']:
        collection = 'landsat-8-l1'
    else:
        raise TypeError(('Satellite "{}" not supported. Use either Landsat-8 or'
                         ' Sentinel-2.').format(satellite))

    # query Development Seed’s API
    r = satsearch.Search.search(intersects=aoi,
                                time='{}/{}'.format(start_date.isoformat(),
                                                    end_date.isoformat()),
                                collection=collection)

    # check if the images footprints contain the area of interest
    aoi = shapely.geometry.shape(aoi)
    results = []
    for x in r.items():
        try:
            if shapely.geometry.shape(x.geometry).contains(aoi):
                results.append(vars(x)['data'])
        except AttributeError:
            pass

    return results


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Search of Landsat-8 and Sentinel-2 images.')
    parser.add_argument('--satellite', choices=['Landsat-8', 'Sentinel-2'],
                        help=('either "Landsat-8" or "Sentinel-2"'),
                        default='Landsat-8')
    parser.add_argument('--geom', type=utils.valid_geojson,
                        help=('path to geojson file'))
    parser.add_argument('--lat', type=utils.valid_lat,
                        help=('latitude of the center of the rectangle AOI'))
    parser.add_argument('--lon', type=utils.valid_lon,
                        help=('longitude of the center of the rectangle AOI'))
    parser.add_argument('-w', '--width', type=int, default=5000,
                        help='width of the AOI (m), default 5000 m')
    parser.add_argument('-l', '--height', type=int, default=5000,
                        help='height of the AOI (m), default 5000 m')
    parser.add_argument('-s', '--start-date', type=utils.valid_datetime,
                        help='start date, YYYY-MM-DD')
    parser.add_argument('-e', '--end-date', type=utils.valid_datetime,
                        help='end date, YYYY-MM-DD')
    args = parser.parse_args()

    if args.geom and (args.lat or args.lon):
        parser.error('--geom and {--lat, --lon} are mutually exclusive')

    if not args.geom and (not args.lat or not args.lon):
        parser.error('either --geom or {--lat, --lon} must be defined')

    if args.geom:
        aoi = args.geom
    else:
        aoi = utils.geojson_geometry_object(args.lat, args.lon, args.width,
                                            args.height)

    print(json.dumps(search(aoi, start_date=args.start_date,
                            end_date=args.end_date, satellite=args.satellite)))



#!/usr/bin/env python3
# vim: set fileencoding=utf-8
# pylint: disable=C0103

"""
Automatic download and crop of Planet images.

Copyright (C) 2016-18, Carlo de Franchis <carlo.de-franchis@m4x.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function
import os
import time
import json
import argparse
import multiprocessing

import area
import requests
import numpy as np
import dateutil.parser
import rasterio

from tsd import utils
from tsd import parallel
from tsd import search_planet

ITEM_TYPES = search_planet.ITEM_TYPES
ASSETS = ['udm',
          'visual',
          'visual_xml',
          'analytic',
          'analytic_xml',
          'analytic_dn',
          'analytic_dn_xml',
          'analytic_sr',
          'basic_udm',
          'basic_analytic',
          'basic_analytic_xml',
          'basic_analytic_rpc',
          'basic_analytic_dn',
          'basic_analytic_dn_xml',
          'basic_analytic_dn_rpc']
client = search_planet.client

CAS_URL = 'https://api.planet.com/compute/ops/clips/v1'  # clip and ship
QUOTA_URL = 'https://api.planet.com/auth/v1/experimental/public/my/subscriptions'


def get_quota():
    """
    Return a string giving the current quota usage.
    """
    r = requests.get(QUOTA_URL, auth=(os.getenv('PL_API_KEY'), ''))
    if r.ok:
        l = r.json()
        #assert(l[0]['plan']['name'] == 'Education and Research Standard (PlanetScope)')
        return '{:.3f} / {} km²'.format(l[0]['quota_used'], l[0]['quota_sqkm'])
    print('ERROR: {} returned {}'.format(QUOTA_URL, r.status_code))


def fname_from_metadata(d):
    """
    Return a string containing the image acquisition date and identifier.

    Args:
        d (dict): dictionary containing a Planet item information
    """
    scene_id = d['id']
    date_str = d['properties']['acquired']
    date = dateutil.parser.parse(date_str).date()
    return '{}_scene_{}'.format(date.isoformat(), scene_id)


def metadata_from_metadata_dict(d):
    """
    Return a dict containing some string-formatted metadata.

    Args:
        d (dict): dictionary containing a Planet item information
    """
    imaging_date = dateutil.parser.parse(d['properties']['acquired'])
    sun_zenith = 90 - d['properties']['sun_elevation']  # zenith and elevation are complementary
    sun_azimuth = d['properties']['sun_azimuth']

    out = {
        "IMAGING_DATE": imaging_date.strftime('%Y-%m-%dT%H:%M:%S'),
        "SUN_ZENITH": str(sun_zenith),
        "SUN_AZIMUTH": str(sun_azimuth)
    }
    out.update({str(k): str(d['properties'][k]) for k in d['properties'].keys()})
    return out


def download_asset(dstfile, asset):
    """
    Download a full asset.
    """
    url = poll_activation(asset)
    if url is not None:
        os.system('wget {} -O {}'.format(url, dstfile))
        #utils.download(url, dstfile)


def download_crop(outfile, asset, aoi, aoi_type):
    """
    Download a crop defined in geographic coordinates using gdal or rasterio.

    Args:
        outfile (string): path to the output file
        asset (dict): dictionary containing the image information
        aoi (geojson.Polygon or 6-tuple): either a (lon, lat) polygon or a UTM
            rectangle (ulx, uly, lrx, lry, utm_zone, lat_band), where
            ulx, uly (floats): x/y UTM coordinates of the upper left (ul) corner
            lrx, lry (floats): x/y UTM coordinates of the lower right (lr) corner
            utm_zone (int): number between 1 and 60 indicating the UTM zone with
                respect to which the UTM coordinates have to be interpreted.
            lat_band (string): letter between C and X indicating the latitude band.
        aoi_type (string): "lonlat_polygon" or "utm_rectangle"
    """
    url = poll_activation(asset)
    if url is not None:
        if aoi_type == "utm_rectangle":
            utils.crop_with_gdal_translate(outfile, url, *aoi)
        elif aoi_type == "lonlat_polygon":
            with rasterio.open(url, 'r') as src:
                rpc_tags = src.tags(ns='RPC')
            crop, x, y = utils.crop_aoi(url, aoi)
            utils.rio_write(outfile, crop,
                            tags={'CROP_OFFSET_XY': '{} {}'.format(x, y)},
                            namespace_tags={'RPC': rpc_tags})


def get_item_asset_info(item, asset_type, verbose=False):
    """
    Get item asset details if we have download permissions.

    Args:
        item (dict): dictionary containing the item info
        asset_type (string): desired asset

    Return:
        (dict): dictionary containing the item asset information
    """
    allowed_assets = client.get_assets(item).get()
    if asset_type not in allowed_assets:
        if verbose:
            print('WARNING: no permission to get asset "{}" of {}'.format(asset_type,
                                                                          item['_links']['_self']))
            print("\tPermissions for this item are:", item['_permissions'])
        return None
    else:
        return allowed_assets[asset_type]


def request_activation(asset):
    """
    Request the activation of an asset to Planet data API.

    Args:
        asset (dict): dictionary containing the image information
    """
    activation = client.activate(asset)
    r = activation.response.status_code
    if r not in [202, 204]:  # 202: activation started
                             # 204: already active
        print('WARNING: activation of asset {} returned {}'.format(asset, r))


def poll_activation(asset):
    """
    Wait for an asset, requested to Planet data API, to be ready for download.

    Args:
        asset (dict): dictionary containing the asset info

    Return:
        (string): url to the file ready for download
    """
    # refresh the asset info
    r = requests.get(asset['_links']['_self'], auth=(os.environ['PL_API_KEY'], ''))
    if r.ok:
        asset = r.json()
    elif r.status_code == 429:  # rate limit
        time.sleep(1)
        return poll_activation(asset)
    else:
        print('ERROR: got {} error code when requesting {}'.format(r.status_code,
                                                                   asset['_links']['_self']))
        return

    # decide what to do next depending on the asset status
    if asset['status'] == 'active':
        return asset['location']
    elif asset['status'] == 'activating':
        time.sleep(3)
        return poll_activation(asset)
    elif asset['status'] == 'inactive':
        request_activation(asset)
        time.sleep(3)
        return poll_activation(asset)
    else:
        print('ERROR: unknown asset status {}'.format(asset['status']))


def request_clip(item, asset, aoi, active=False):
    """
    Request a clip to Planet clip & ship API.

    Args:
        item (dict): dictionary containing the item info
        asset (dict): dictionary containing the asset info
        aoi (dict): dictionary containing a geojson polygon (e.g. output of
            utils.geojson_geometry_object)
        active (bool): boolean

    Return:
        (dict): dictionary containing the clip info
    """
    if not active:  # wait for the asset to be actived
        poll_activation(asset)

    # request the clip
    d = {
        "aoi": aoi,
        "targets": [
            {
                "item_id": item['id'],
                "item_type": item['properties']['item_type'],
                "asset_type": asset['type']
            }
        ]
    }
    headers = {'content-type': 'application/json'}
    r = requests.post(CAS_URL,
                      headers=headers, data=json.dumps(d),
                      auth=(os.environ['PL_API_KEY'], ''))
    if r.ok:
        return r.json()
    elif r.status_code == 429:  # rate limit
        time.sleep(1)
        return request_clip(item, asset, aoi, active=True)
    else:
        print('ERROR: got {} error code when requesting {}'.format(r.status_code, d))


def poll_clip(clip_json):
    """
    Wait for a clip, requested to Planet clip & ship API, to be ready.

    Args:
        clip_json (dict): dictionary containing the clip info

    Return:
        (string): url to the zipfile containing the clipped data.
    """
    # refresh the clip info
    clip_request_url = clip_json['_links']['_self']
    r = requests.get(clip_request_url, auth=(os.environ['PL_API_KEY'], ''))
    if r.ok:
        j = r.json()
    elif r.status_code == 429:  # rate limit
        time.sleep(1)
        return poll_clip(clip_json)
    else:
        print('ERROR: got {} error code when requesting {}'.format(r.status_code, clip_request_url))
        return

    # decide what to do next depending on the clip status
    if j['state'] == 'succeeded':
        return j['_links']['results'][0]
    elif j['state'] == 'running':
        time.sleep(3)
        return poll_clip(clip_json)
    else:
        print('ERROR: unknown state "{}" of clip request {}'.format(j['state'],
                                                                    clip_request_url))


def download_clip(clip_info, outpath):
    """
    Download a zipfile from Planet clip & ship endpoint after a clip request.

    Args:
        clip_info (dict): dictionary containing the clip info
        outpath (string): path where to store the downloaded zip file
    """
    url = poll_clip(clip_info)
    utils.download(url, outpath, auth=(os.environ['PL_API_KEY'], ''))


def get_time_series(aoi, start_date=None, end_date=None,
                    item_types=['PSScene3Band'], asset_type='analytic',
                    out_dir='',
                    parallel_downloads=multiprocessing.cpu_count(),
                    clip_and_ship=True, no_crop=False, satellite_id=None,
                    search_type='contains', remove_duplicates=True):
    """
    Main function: crop and download Planet images.
    """
    # list available images
    items = search_planet.search(aoi, start_date, end_date,
                                 item_types=item_types,
                                 satellite_id=satellite_id,
                                 search_type=search_type,
                                 remove_duplicates=remove_duplicates)
    print('Found {} images'.format(len(items)))

    # list the requested asset for each available (and allowed) image
    print('Listing available {} assets...'.format(asset_type), flush=True, end=' ')
    assets = parallel.run_calls(get_item_asset_info, items,
                                extra_args=(asset_type,), pool_type='threads',
                                nb_workers=parallel_downloads, timeout=600)

    # remove 'None' (ie not allowed) assets and corresponding items
    items = [i for (i, a) in zip(items, assets) if a]
    assets = [a for a in assets if a]
    print('Have permissions for {} images'.format(len(items)))

    # activate the allowed assets
    print('Requesting activation of {} images...'.format(len(assets)),
          flush=True, end=' ')
    parallel.run_calls(request_activation, assets, pool_type='threads',
                       nb_workers=parallel_downloads, timeout=600)

    # warn user about quota usage
    n = len(assets)
    if clip_and_ship:
        a = n * area.area(aoi)
    else:
        a = np.sum(area.area(i['geometry']) for i in items)
    print('Your current quota usage is {}'.format(get_quota()), flush=True)
    print('Downloading these {} images will increase it by {:.3f} km²'.format(n, a/1e6),
          flush=True)

    # build filenames
    ext = 'zip' if clip_and_ship else 'tif'
    out_dir = os.path.abspath(os.path.expanduser(out_dir))
    fnames = [os.path.join(out_dir, '{}.{}'.format(fname_from_metadata(i),
                                                   ext)) for i in items]

    if clip_and_ship:
        print('Requesting clip of {} images...'.format(len(assets)),
              flush=True, end=' ')
        clips = parallel.run_calls(request_clip, list(zip(items, assets)),
                                   extra_args=(aoi,), pool_type='threads',
                                   nb_workers=parallel_downloads, timeout=3600)

        # remove clips that were rejected
        ok = [i for i, x in enumerate(clips) if x]
        clips = [clips[i] for i in range(len(clips)) if i in ok]
        fnames = [fnames[i] for i in range(len(fnames)) if i in ok]

        print('Downloading {} clips...'.format(len(clips)), end=' ', flush=True)
        parallel.run_calls(download_clip, list(zip(clips, fnames)),
                           pool_type='threads', nb_workers=parallel_downloads,
                           timeout=3600)

    elif no_crop:  # download full images
        os.makedirs(out_dir, exist_ok=True)
        print('Downloading {} full images...'.format(len(assets)), end=' ')
        parallel.run_calls(download_asset, list(zip(fnames, assets)),
                           pool_type='threads', nb_workers=parallel_downloads,
                           timeout=1200)
    else:
        if asset_type in ['udm', 'visual', 'analytic', 'analytic_dn',
                          'analytic_sr']:
            aoi_type = 'utm_rectangle'
            aoi = utils.utm_bbx(aoi)
        else:
            aoi_type = 'lonlat_polygon'

        # download crops with gdal through vsicurl
        os.makedirs(out_dir, exist_ok=True)
        print('Downloading {} crops...'.format(len(assets)), end=' ')
        parallel.run_calls(download_crop, list(zip(fnames, assets)),
                           extra_args=(aoi, aoi_type),
                           pool_type='threads', nb_workers=parallel_downloads,
                           timeout=300)

        # embed some metadata in the image files
        for f, img in zip(fnames, items):  # embed some metadata as gdal geotiff tags
            if os.path.isfile(f):
                utils.set_geotif_metadata_items(f, metadata_from_metadata_dict(img))


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description=('Automatic download and crop '
                                                  'of Planet images'))
    parser.add_argument('--geom', type=utils.valid_geojson,
                        help=('path to geojson file'))
    parser.add_argument('--lat', type=utils.valid_lat,
                        help=('latitude of the center of the rectangle AOI'))
    parser.add_argument('--lon', type=utils.valid_lon,
                        help=('longitude of the center of the rectangle AOI'))
    parser.add_argument('-w', '--width', type=int, default=2000,
                        help='width of the AOI (m), default 2000 m')
    parser.add_argument('-l', '--height', type=int, default=2000,
                        help='height of the AOI (m), default 2000 m')
    parser.add_argument('-s', '--start-date', type=utils.valid_datetime,
                        help='start date, YYYY-MM-DD')
    parser.add_argument('-e', '--end-date', type=utils.valid_datetime,
                        help='end date, YYYY-MM-DD')
    parser.add_argument('--search-type', choices=['contains', 'intersects'],
                        default='contains', help='search type')
    parser.add_argument('--satellite-id', help='satellite identifier, e.g. 0f02')
    parser.add_argument('--keep-duplicates', action='store_true',
                        help='keep all images even when two were acquired within'
                             ' less than 5 minutes (the default behaviour is to'
                             ' discard such duplicates)')
    parser.add_argument('--item-types', nargs='*', choices=ITEM_TYPES,
                        default=['PSScene3Band'], metavar='',
                        help=('space separated list of item types to'
                              ' download. Default is PSScene3Band. Allowed'
                              ' values are {}'.format(', '.join(ITEM_TYPES))))
    parser.add_argument('--asset', default='analytic', metavar='',
                        choices=ASSETS,
                        help=('asset item type to download. Default is analytic.'
                              ' Allowed values are {}'.format(', '.join(ASSETS))))
    parser.add_argument('-o', '--outdir', type=str, help=('path to save the '
                                                          'images'), default='')
    parser.add_argument('--parallel-downloads', type=int, default=10,
                        help='max number of parallel crops downloads')
    parser.add_argument('--clip-and-ship', action='store_true', help=('use the '
                                                                      'clip and '
                                                                      'ship API'))
    parser.add_argument('--no-crop', action='store_true',
                        help=("don't crop but instead download the whole image files"))
    args = parser.parse_args()

    if args.geom and (args.lat or args.lon):
        parser.error('--geom and {--lat, --lon} are mutually exclusive')

    if args.clip_and_ship and args.no_crop:
        parser.error('--clip-and-ship and --no-crop are mutually exclusive')

    if not args.geom and (not args.lat or not args.lon):
        parser.error('either --geom or {--lat, --lon} must be defined')

    if args.geom:
        aoi = args.geom
    else:
        aoi = utils.geojson_geometry_object(args.lat, args.lon, args.width,
                                            args.height)
    get_time_series(aoi, start_date=args.start_date, end_date=args.end_date,
                    item_types=args.item_types, asset_type=args.asset,
                    out_dir=args.outdir,
                    parallel_downloads=args.parallel_downloads,
                    clip_and_ship=args.clip_and_ship,
                    no_crop=args.no_crop,
                    search_type=args.search_type,
                    satellite_id=args.satellite_id,
                    remove_duplicates=~args.keep_duplicates)



from tsd import utils
from tsd import search_devseed
from tsd import search_scihub
from tsd import search_planet
from tsd import s2_metadata_parser
from tsd import l8_metadata_parser

from tsd import get_landsat
from tsd import get_sentinel2
from tsd import get_sentinel1
from tsd import get_planet



"""
This module contains parsers for the Sentinel-2 metadata outputs of all the
search APIs supported by TSD, such as devseed, planet, scihub and gcloud. Each
API parser receives as input a Python dict containing the metadata of an image
as returned by the API. It extracts from it the metadata that TSD needs and
stores them in an object with standard attributes (i.e. the attributes are the
same for all APIs). The detailed list of attributes is given below. This allows
TSD to use any search API with any download mirror.

Each parser returns a Sentinel2Images object with the following attributes:

    utm_zone (int): integer between 1 and 60 indicating the UTM longitude zone
    lat_band (str): letter between C and X, excluding I and O, indicating the
        UTM latitude band
    sqid (str): pair of letters indicating the MGRS 100x100 km square
    mgrd_id (str): concatenation of utm_zone, lat_band and sqid. It has length
        five (utm_zone is zero padded).
    date (datetime.datetime): acquisition date and time of the image
    satellite (str): either 'S2A' or 'S2B'
    orbit (int): relative orbit number
    title (str): original name of the SAFE in which the image is packaged by ESA
    filename (str): string that TSD uses to name the crops downloaded for the bands
        of this image. It starts with the acquisition year, month and day so that
        sorting the files per image acquisition date is easy.
    urls (dict): dict with keys 'aws' and 'gcloud'. The value associated to
        each key is a dict with one key per band containing download urls.
    metadata_original (dict): the original response of the API for this image
"""
from __future__ import print_function
import re
import datetime
import dateutil.parser
import requests

from tsd import search_scihub

AWS_S3_URL_L1C = 's3://sentinel-s2-l1c'
AWS_S3_URL_L2A = 's3://sentinel-s2-l2a'
GCLOUD_URL_L1C = 'https://storage.googleapis.com/gcp-public-data-sentinel-2'
SCIHUB_API_URL = 'https://scihub.copernicus.eu/apihub/odata/v1'
RODA_URL = 'https://roda.sentinel-hub.com/sentinel-s2-l1c/tiles'

ALL_BANDS = ['TCI', 'B01', 'B02', 'B03', 'B04', 'B05', 'B06', 'B07', 'B08',
             'B8A', 'B09', 'B10', 'B11', 'B12']

BANDS_RESOLUTION = {'B01': 60,
                    'B02': 10,
                    'B03': 10,
                    'B04': 10,
                    'B05': 20,
                    'B06': 20,
                    'B07': 20,
                    'B08': 10,
                    'B8A': 20,
                    'B09': 60,
                    'B10': 60,
                    'B11': 20,
                    'B12': 20}


def split_mgrs_id(mgrs_id):
    """
    Split an mgrs identifier such as 10SEG into (10, S, EG).
    """
    return re.split('(\d+)([a-zA-Z])([a-zA-Z]+)', mgrs_id)[1:4]


def parse_safe_name_for_relative_orbit_number(safe_name):
    """
    """
    s = re.search('_R([0-9]{3})_', safe_name)
    return int(s.group(1))


def parse_safe_name_for_mgrs_id(safe_name):
    """
    """
    return re.findall(r"_T([0-9]{2}[A-Z]{3})_", safe_name)[0]


def parse_datastrip_id_for_granule_date(datastrip_id):
    """
    Parse a datastrip id for the corresponding granule acquisition date.

    Examples of datastrip ids:
      S2B_OPER_MSI_L1C_DS_SGS__20180510T205109_S20180510T185438_N02.06 -> 20180510T185438
      S2A_OPER_MSI_L1C_DS_EPAE_20180516T000159_S20180515T190003_N02.06 -> 20180515T190003
    """
    date_str = re.findall(r"_S(2[0-9]{3}[0-1][0-9][0-3][0-9]T[0-9]{6})_",
                          datastrip_id)[0]
    return dateutil.parser.parse(date_str)


def parse_datatake_id_for_absolute_orbit(datatake_id):
    """
    Examples of datatake ids:
        GS2B_20180510T184929_006145_N02.06
        GS2A_20180515T184941_015125_N02.06
    """
    return int(datatake_id.split('_')[2])


def filename_from_metadata(img):
    """
    Args:
        img (Sentinel2Image instance): Sentinel-2 image metadata
    """
    return '{}_{}_orbit_{:03d}_tile_{}'.format(img.date.date().isoformat(),
                                               img.satellite,
                                               img.relative_orbit,
                                               img.mgrs_id)


def get_s2_granule_id_of_scihub_item_from_scihub(img):
    """
    Build the granule id of a given single tile SAFE.

    The hard part is to get the timestamp in the granule id. Unfortunately this
    timestamp is not part of the metadata returned by scihub. This function queries
    scihub OData API to retrieve it. It can be insanely slow (a few minutes).
    Another con is that it needs credentials.

    Args:
        img (dict): single SAFE metadata as returned by scihub opensearch API

    Return:
        str: granule id, e.g. L1C_T36RTV_A005095_20180226T084545
    """
    granule_request = "{}/Products('{}')/Nodes('{}')/Nodes('GRANULE')/Nodes?$format=json".format(SCIHUB_API_URL,
                                                                                                 img['id'],
                                                                                                 img['filename'])
    granules = requests.get(granule_request,
                            auth=(search_scihub.read_copernicus_credentials_from_environment_variables())).json()
    return granules["d"]["results"][0]["Id"]


def get_s2_granule_id_of_scihub_item_from_sentinelhub(img):
    """
    Build the granule id of a given single tile SAFE.

    The hard part is to get the timestamp in the granule id. Unfortunately this
    timestamp is not part of the metadata returned by scihub. This function queries
    sentinelhub to retrieve it. It takes about 3 seconds.

    Args:
        img (Sentinel2Image instance): Sentinel-2 image metadata

    Return:
        str: granule id, e.g. L1C_T36RTV_A005095_20180226T084545
    """
    import sentinelhub
    t0 = (img.date - datetime.timedelta(hours=2)).isoformat()
    t1 = (img.date + datetime.timedelta(hours=2)).isoformat()
    r = sentinelhub.opensearch.get_tile_info('T{}'.format(img.mgrs_id), time=(t0, t1))
    assert(isinstance(r, dict))

    granule_date = dateutil.parser.parse(r['properties']['startDate']).strftime("%Y%m%dT%H%M%S")
    return "L1C_T{}_A{:06d}_{}".format(img.mgrs_id, img.relative_orbit, granule_date)


def get_roda_metadata(img, filename='tileInfo.json'):
    """
    Args:
        img (Sentinel2Image instance): Sentinel-2 image metadata

    Return:
        dict: content of the roda metadata json file
    """
    url = '{}/{}/{}/{}/{}/{}/{}/0/{}'.format(RODA_URL, img.utm_zone,
                                             img.lat_band, img.sqid,
                                             img.date.year, img.date.month,
                                             img.date.day, filename)
    r = requests.get(url)
    if r.ok:
        return r.json()
    else:
        raise Exception("{} not found on roda".format(img.title))


class Sentinel2Image():
    """
    """
    def __init__(self, img, api='devseed'):
        """
        """
        self.metadata_source = api
        #self.metadata_original = img

        if api == 'devseed':
            self.devseed_parser(img)
        elif api == 'scihub':
            self.scihub_parser(img)
        elif api == 'planet':
            self.planet_parser(img)
        elif api == 'gcloud':
            self.gcloud_parser(img)

        self.filename = filename_from_metadata(self)
        self.urls = {'aws': {}, 'gcloud': {}}


    def devseed_parser(self, img):
        """
        Args:
            img (dict): json metadata dict as shipped in devseed API response
        """
        p = img['properties']
        self.title = p['sentinel:product_id']
        self.utm_zone = p['sentinel:utm_zone']
        self.lat_band = p['sentinel:latitude_band']
        self.sqid  = p['sentinel:grid_square']
        self.mgrs_id = '{}{}{}'.format(self.utm_zone, self.lat_band, self.sqid)

        self.date = dateutil.parser.parse(self.title.split('_')[2])
        #self.granule_date = dateutil.parser.parse(p['datetime'])
        self.satellite = p['eo:platform'].replace("sentinel-", "S").upper()  # sentinel-2b --> S2B
        self.relative_orbit = parse_safe_name_for_relative_orbit_number(self.title)

        self.thumbnail = img['assets']['thumbnail']['href'].replace('sentinel-s2-l1c.s3.amazonaws.com',
                                                                    'roda.sentinel-hub.com/sentinel-s2-l1c')
        self.cloud_cover = p['eo:cloud_cover']
        #self.id = img['id']


    def scihub_parser(self, img):
        """
        Args:
            img (dict): json metadata dict for a single SAFE, as shipped in scihub
                opensearch API response
        """
        self.title = img['title']
        try:
            self.mgrs_id = img['tileid']
        except KeyError:
            self.mgrs_id = re.findall(r"_T([0-9]{2}[A-Z]{3})_", img['title'])[0]
        self.utm_zone, self.lat_band, self.sqid = split_mgrs_id(self.mgrs_id)
        self.date = dateutil.parser.parse(img['beginposition'], ignoretz=True)
        self.satellite = self.title[:3]  # S2A_MSIL1C_2018010... --> S2A
        self.absolute_orbit = img['orbitnumber']
        self.relative_orbit = img['relativeorbitnumber']
        self.thumbnail = img['links']['icon']


    def planet_parser(self, img):
        """
        Args:
            img (dict): json metadata dict for a single SAFE, as shipped in Planet
                API response
        """
        self.title = img['id']
        self.mgrs_id = img['properties']['mgrs_grid_id']
        self.utm_zone, self.lat_band, self.sqid = split_mgrs_id(self.mgrs_id)
        self.date = dateutil.parser.parse(img['properties']['acquired'])
        self.satellite = img['properties']['satellite_id'].replace("Sentinel-", "S")  # Sentinel-2A --> S2A
        self.relative_orbit = img['properties']['rel_orbit_number']
        self.absolute_orbit = None ## TODO
        self.thumbnail = None  ## TODO


    def gcloud_parser(self, img):
        """
        Args:
            img (dict): json metadata dict for a single SAFE, as shipped in Gcloud
                API response
        """
        self.title = img['product_id']
        self.mgrs_id = img['mgrs_tile']
        self.utm_zone, self.lat_band, self.sqid = split_mgrs_id(self.mgrs_id)
        self.date = dateutil.parser.parse(img['sensing_time'], ignoretz=True)
        self.satellite = img['product_id'][:3]
        self.relative_orbit = parse_safe_name_for_relative_orbit_number(self.title)

        self.absolute_orbit = int(img['granule_id'].split('_')[2][1:])
        self.granule_date = dateutil.parser.parse(img['granule_id'].split('_')[3])

        self.thumbnail = None  ## TODO
        #self.is_old = True if '.' in img['granule_id'] else False


    def build_gs_links(self):
        """
        Build Gcloud urls for the 13 jp2 bands and the gml cloud mask.

        Example of url:
        https://storage.googleapis.com/gcp-public-data-sentinel-2/tiles/36/R/TV/S2B_MSIL1C_20180226T083909_N0206_R064_T36RTV_20180226T122942.SAFE/GRANULE/L1C_T36RTV_A005095_20180226T084545/IMG_DATA/T36RTV_20180226T083909_B01.jp2

        The tricky part is to build the granule name
        (L1C_T36RTV_A005095_20180226T084545 in the example above), which is not
        part neither of the devseed nor of the scihub API responses. This function
        queries roda to retrieve it. It takes about 200 ms.
        """
        if not hasattr(self, 'granule_date'):
            tile_info = get_roda_metadata(self, filename='tileInfo.json')
            #self.granule_date = dateutil.parser.parse(tile_info['timestamp'])
            self.granule_date = parse_datastrip_id_for_granule_date(tile_info['datastrip']['id'])

        if not hasattr(self, 'absolute_orbit'):
            product_info = get_roda_metadata(self, filename='productInfo.json')
            self.absolute_orbit = parse_datatake_id_for_absolute_orbit(product_info['datatakeIdentifier'])

    #    if self.is_old:
    #        img_name = '{}_{}.jp2'.format('_'.join(granule_id.split('_')[:-1]), '{}')
    #        cloud_mask_name = '{}_B00_MSIL1C.gml'.format('_'.join(granule_id.split('_')[:-1]).replace('MSI_L1C_TL', 'MSK_CLOUDS'))

        granule_id = 'L1C_T{}_A{:06d}_{}'.format(self.mgrs_id,
                                                 self.absolute_orbit,
                                                 self.granule_date.strftime("%Y%m%dT%H%M%S"))
        base_url = '{}/tiles/{}/{}/{}/{}.SAFE/GRANULE/{}'.format(GCLOUD_URL_L1C,
                                                                 self.utm_zone,
                                                                 self.lat_band,
                                                                 self.sqid,
                                                                 self.title,
                                                                 granule_id)
        urls = self.urls['gcloud']
        urls['cloud_mask'] = '{}/QI_DATA/MSK_CLOUDS_B00.gml'.format(base_url)
        for b in ALL_BANDS:
            urls[b] = '{}/IMG_DATA/T{}_{}_{}.jp2'.format(base_url,
                                                         self.mgrs_id,
                                                         self.date.strftime("%Y%m%dT%H%M%S"),
                                                         b)


    def build_s3_links(self):
        """
        Build s3 urls for the 13 jp2 bands and the gml cloud mask.

        Example of url: s3://sentinel-s2-l1c/tiles/10/S/EG/2018/2/24/0/B04.jp2
        """
        aws_s3_url = AWS_S3_URL_L2A if 'MSIL2A' in self.title else AWS_S3_URL_L1C
        base_url = '{}/tiles/{}/{}/{}/{}/{}/{}/0'.format(aws_s3_url,
                                                         self.utm_zone,
                                                         self.lat_band,
                                                         self.sqid,
                                                         self.date.year,
                                                         self.date.month,
                                                         self.date.day)
        urls = self.urls['aws']
        urls['cloud_mask'] = '{}/qi/MSK_CLOUDS_B00.gml'.format(base_url)
        for b in ALL_BANDS:
            if 'MSIL2A' in self.title:
                urls[b] = '{}/R{}m/{}.jp2'.format(base_url, BANDS_RESOLUTION[b], b)
            else:
                urls[b] = '{}/{}.jp2'.format(base_url, b)



VERSION = (
    0,
    4,
    11,
)

__title__ = "tsd"
__description__ = "Search and download of Sentinel, Landsat and Planet crops."
__url__ = 'https://github.com/cmla/tsd'
__version__ = ".".join(map(str, VERSION))
__author__ = "Carlo de Franchis"
__author_email__ = 'carlo.de-franchis@ens-cachan.fr'



#!/usr/bin/env python3
# vim: set fileencoding=utf-8
# pylint: disable=C0103

"""
Automatic download of Sentinel-1 images.

Copyright (C) 2016-18, Carlo de Franchis <carlo.de-franchis@ens-cachan.fr>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from __future__ import print_function
import os
import zipfile
import argparse
import subprocess
import dateutil.parser

import bs4
import requests

from tsd import utils
from tsd import search_scihub


PEPS_URL_SEARCH = 'https://peps.cnes.fr/resto/api/collections'
PEPS_URL_DOWNLOAD = 'https://peps.cnes.fr/resto/collections'
CODEDE_URL = 'https://code-de.org/Sentinel1'


def read_copernicus_credentials_from_environment_variables():
    """
    Read the user Copernicus Open Access Hub credentials.
    """
    try:
        login = os.environ['COPERNICUS_LOGIN']
        password = os.environ['COPERNICUS_PASSWORD']
    except KeyError as e:
        print("The {} module requires the COPERNICUS_LOGIN and".format(os.path.basename(__file__)),
              "COPERNICUS_PASSWORD environment variables to be defined with valid",
              "credentials for https://scihub.copernicus.eu/. Create an account if",
              "you don't have one (it's free) then edit the relevant configuration",
              "files (eg .bashrc) to define these environment variables.")
        raise e
    return login, password


def query_data_hub(output_filename, url, user, password, verbose=False):
    """
    Download a file from the Copernicus data hub or one of its true mirrors.
    """
    verbosity = '--verbose' if verbose else '--no-verbose'  # intermediate verbosity with --quiet
    subprocess.call(['wget',
                     verbosity,
                     '--no-check-certificate',
                     '--auth-no-challenge',
                     '--user=%s' % user,
                     '--password=%s' % password,
                     '--output-document=%s' % output_filename,
                     url])


def download_safe_from_peps(safe_name, out_dir=''):
    """
    """
    try:
        login, password = os.environ['PEPS_LOGIN'], os.environ['PEPS_PASSWORD']
    except KeyError:
        print("Downloading from PEPS requires the PEPS_LOGIN and".format(__file__),
              "PEPS_PASSWORD environment variables to be defined with valid",
              "credentials for https://peps.cnes.fr/. Create an account if",
              "you don't have one (it's free) then edit the relevant configuration",
              "files (eg .bashrc) to define these environment variables.")
        return

    query = '{}/S1/search.atom?identifier={}'.format(PEPS_URL_SEARCH, safe_name)
    r = requests.get(query)

    if not r.ok:
        print('WARNING: request {} failed'.format(query))
        return

    img = bs4.BeautifulSoup(r.text, 'xml').find_all('entry')[0]
    peps_id = img.find('id').text
    url = "{}/S1/{}/download".format(PEPS_URL_DOWNLOAD, peps_id)
    zip_path = os.path.join(out_dir, '{}.SAFE.zip'.format(safe_name))
    cmd = "curl -k --basic -u {}:{} {} -o {}".format(login, password,
                                                     url, zip_path)
    print(cmd)
    os.system(cmd)


def download_sentinel_image(image, out_dir='', mirror='peps'):
    """
    Download a Sentinel image.
    """
    # create output directory
    if out_dir:
        out_dir = os.path.abspath(os.path.expanduser(out_dir))
        os.makedirs(out_dir, exist_ok=True)

    # download zip file
    zip_path = os.path.join(out_dir, '{}.SAFE.zip'.format(image['title']))
    date = dateutil.parser.parse(image['beginposition'], ignoretz=True)
    if not zipfile.is_zipfile(zip_path) or os.stat(zip_path).st_size == 0:
        if mirror == 'code-de':
            url = '{}/{:04d}/{:02d}/{:02d}/{}.SAFE.zip'.format(CODEDE_URL,
                                                               date.year,
                                                               date.month,
                                                               date.day,
                                                               image['title'])
            if requests.head(url).ok:  # download the file
                subprocess.call(['wget', url])
            else:  # switch to PEPS
                print('WARNING: {} not available, trying from PEPS...'.format(url))
                download_sentinel_image(image, out_dir, mirror='peps')
        elif mirror == 'peps':
            try:
                download_safe_from_peps(image['title'], out_dir=out_dir)
            except Exception:
                print('WARNING: failed request to {}/S1/search.atom?identifier={}'.format(PEPS_URL_SEARCH, image['title']))
                print('WARNING: will download from copernicus mirror...')
                download_sentinel_image(image, out_dir, mirror='copernicus')
        elif mirror in search_scihub.API_URLS:
            url = "{}/odata/v1/Products('{}')/$value".format(search_scihub.API_URLS[mirror],
                                                             image['id'])
            user, password = read_copernicus_credentials_from_environment_variables()
            query_data_hub(zip_path, url, user, password, verbose=False)
        else:
            print('ERROR: unknown mirror {}'.format(mirror))

    return zip_path


def get_time_series(aoi, start_date=None, end_date=None, out_dir='',
                    product_type='GRD', operational_mode='IW',
                    relative_orbit_number=None, swath_identifier=None,
                    search_api='copernicus', download_mirror='peps'):
    """
    Main function: download a Sentinel-1 image time serie.
    """
    # list available images
    images = search_scihub.search(aoi, start_date, end_date,
                                  product_type=product_type,
                                  operational_mode=operational_mode,
                                  swath_identifier=swath_identifier,
                                  relative_orbit_number=relative_orbit_number,
                                  api=search_api)

    # download
    for image in images:
        download_sentinel_image(image, out_dir, download_mirror)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description=('Automatic download '
                                                  'of Sentinel-1 images'))
    parser.add_argument('--geom', type=utils.valid_geojson,
                        help=('path to geojson file'))
    parser.add_argument('--lat', type=utils.valid_lat,
                        help=('latitude of the center of the rectangle AOI'))
    parser.add_argument('--lon', type=utils.valid_lon,
                        help=('longitude of the center of the rectangle AOI'))
    parser.add_argument('-w', '--width', type=int, default=5000,
                        help='width of the AOI (m), default 5000 m')
    parser.add_argument('-l', '--height', type=int, default=5000,
                        help='height of the AOI (m), default 5000 m')
    parser.add_argument('-s', '--start-date', type=utils.valid_datetime,
                        help='start date, YYYY-MM-DD')
    parser.add_argument('-e', '--end-date', type=utils.valid_datetime,
                        help='end date, YYYY-MM-DD')
    parser.add_argument('--title', help='image title (filename.SAFE)')
    parser.add_argument('-o', '--outdir', type=str, help=('path to save the '
                                                          'images'), default='')
    parser.add_argument('-t', '--product-type',
                        help='type of image: GRD, SLC, RAW', default='GRD')
    parser.add_argument('-m', '--operational-mode', default='IW',
                        help='acquisiton mode: SM, IW, EW or WV')
    parser.add_argument('--swath-identifier',
                        help='(for S1) subswath id: S1..S6 or IW1..IW3 or EW1..EW5')
    parser.add_argument('--api', default='copernicus',
                        help='search API to use: copernicus, austria or finland')
    parser.add_argument('--mirror', default='peps',
                        help='download mirror: peps, copernicus, austria or finland')
    args = parser.parse_args()

    if args.geom and (args.lat or args.lon):
        parser.error('--geom and {--lat, --lon} are mutually exclusive')

    if args.title and (args.geom or args.lat or args.lon):
        parser.error('--title, --geom and {--lat, --lon} are mutually exclusive')

    if not args.geom and (not args.lat or not args.lon) and (not args.title):
        parser.error('either --geom, {--lat, --lon} or --title must be defined')

    if args.title:
        download_safe_from_peps(args.title, args.outdir)
    else:
        if args.geom:
            aoi = args.geom
        else:
            aoi = utils.geojson_geometry_object(args.lat, args.lon, args.width,
                                                args.height)
        get_time_series(aoi, start_date=args.start_date, end_date=args.end_date,
                        out_dir=args.outdir, product_type=args.product_type,
                        operational_mode=args.operational_mode,
                        swath_identifier=args.swath_identifier,
                        search_api=args.api, download_mirror=args.mirror)



import datetime

from tsd import search_scihub
from tsd import utils

aoi = utils.geojson_geometry_object(29.9793, 31.1346, 5000, 5000)
results = search_scihub.search(aoi,
                               start_date=datetime.datetime(2019, 1, 1),
                               end_date=datetime.datetime(2019, 1, 15),
                               satellite='Sentinel-2')
expected_titles = ['S2A_MSIL1C_20190114T083311_N0207_R021_T36RUU_20190114T085705',
                   'S2B_MSIL1C_20190109T083329_N0207_R021_T36RUU_20190109T103019',
                   'S2A_MSIL1C_20190104T083331_N0207_R021_T36RUU_20190104T104619']
assert([r['title'] for r in results] == expected_titles)



#!/usr/bin/env python3
# vim: set fileencoding=utf-8
# pylint: disable=C0103

"""
Automatic crop and download of Sentinel-2 images.

Copyright (C) 2016-18, Carlo de Franchis <carlo.de-franchis@m4x.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function
import re
import os
import sys
import shutil
import argparse
import multiprocessing
import dateutil.parser
import datetime
import requests
import bs4
import boto3
import botocore
import geojson
import shapely.geometry

from tsd import utils
from tsd import parallel
from tsd import s2_metadata_parser

# list of spectral bands
ALL_BANDS = ['TCI', 'B01', 'B02', 'B03', 'B04', 'B05', 'B06', 'B07', 'B08',
             'B8A', 'B09', 'B10', 'B11', 'B12']


def check_args(api, mirror, product_type):
    if product_type is not None and api != 'scihub':
        print("WARNING: product_type option is available only with api='scihub'")
    if mirror == 'gcloud' and api not in ['gcloud', 'devseed', 'scihub']:
        raise ValueError(
            "ERROR: You must use gcloud or devseed api to use gcloud as mirror")
    if api == 'gcloud' and mirror != 'gcloud':
        raise ValueError(
            "ERROR: You must use gcloud mirror to use gcloud as api")
    if api == 'gcloud':
        try:
            private_key = os.environ['GOOGLE_APPLICATION_CREDENTIALS']
        except KeyError:
            raise ValueError(
                'You must have the env variable GOOGLE_APPLICATION_CREDENTIALS linking to the cred json file')
    if mirror == 'aws':
        info_url = "https://forum.sentinel-hub.com/t/changes-of-the-access-rights-to-l1c-bucket-at-aws-public-datasets-requester-pays/172"

        if 'AWS_ACCESS_KEY_ID' in os.environ and 'AWS_SECRET_ACCESS_KEY' in os.environ:
            try:
                boto3.session.Session().client('s3').list_objects_v2(Bucket=s2_metadata_parser.AWS_S3_URL_L1C[5:],
                                                                     RequestPayer='requester')
            except botocore.exceptions.ClientError:
                raise ValueError(
                    'Could not connect to AWS server. Check credentials or use mirror=gcloud')
        else:
            raise ValueError("TSD downloads Sentinel-2 image crops from the s3://sentinel-s2-l1c",
                             "AWS bucket, which used to be free. On the 7th of August 2018,",
                             "the bucket was switched to 'Requester Pays'. As a consequence,",
                             "you need an AWS account and your credentials stored in the",
                             "AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY environment",
                             "variables in order to use TSD. The price ranges in 0.05-0.09 $",
                             "per GB. More info: {}".format(info_url))


def search(aoi, start_date=None, end_date=None, product_type=None,
           api='devseed'):
    """
    Search Sentinel-2 images covering an AOI and timespan using a given API.

    Args:
        aoi (geojson.Polygon): area of interest
        start_date (datetime.datetime): start of the search time range
        end_date (datetime.datetime): end of the search time range
        product_type (str, optional): either 'L1C' or 'L2A'
        api (str, optional): either devseed (default), scihub, planet or gcloud

    Returns:
        list of image objects
    """
    # list available images
    if api == 'devseed':
        from tsd import search_devseed
        images = search_devseed.search(aoi, start_date, end_date,
                                       'Sentinel-2')
    elif api == 'scihub':
        from tsd import search_scihub
        if product_type is not None:
            product_type = 'S2MSI{}'.format(product_type[1:])
        images = search_scihub.search(aoi, start_date, end_date,
                                      satellite='Sentinel-2',
                                      product_type=product_type)
    elif api == 'planet':
        from tsd import search_planet
        images = search_planet.search(aoi, start_date, end_date,
                                      item_types=['Sentinel2L1C'])
    elif api == 'gcloud':
        from tsd import search_gcloud
        images = search_gcloud.search(aoi, start_date, end_date)

    # parse the API metadata
    images = [s2_metadata_parser.Sentinel2Image(img, api) for img in images]

    # sort images by acquisition day, then by mgrs id
    images.sort(key=(lambda k: (k.date.date(), k.mgrs_id)))

    # remove duplicates (same acquisition date but different mgrs tile id)
    seen = set()
    unique_images = []
    for img in images:
        if img.date not in seen:
            seen.add(img.date)
            unique_images.append(img)

    print('Found {} images'.format(len(unique_images)))
    return unique_images


def download(imgs, bands, aoi, mirror, out_dir, parallel_downloads):
    """
    Download a timeseries of crops with GDAL VSI feature.

    Args:
        imgs (list): list of images
        bands (list): list of bands
        aoi (geojson.Polygon): area of interest
        mirror (str): either 'aws' or 'gcloud'
        out_dir (str): path where to store the downloaded crops
        parallel_downloads (int): number of parallel downloads
    """
    print('Building {} {} download urls...'.format(len(imgs), mirror),)
    if mirror == 'gcloud':
        parallel.run_calls(s2_metadata_parser.Sentinel2Image.build_gs_links,
                           imgs, pool_type='threads',
                           nb_workers=parallel_downloads)
    else:
        parallel.run_calls(s2_metadata_parser.Sentinel2Image.build_s3_links,
                           imgs, pool_type='threads',
                           nb_workers=parallel_downloads)

    crops_args = []
    for img in imgs:
        # convert aoi coords from (lon, lat) to UTM in the zone of the image
        coords = utils.utm_bbx(aoi, utm_zone=int(img.utm_zone),
                               r=60)  # round to multiples of 60 (B01 resolution)

        for b in bands:
            fname = os.path.join(out_dir, '{}_band_{}.tif'.format(img.filename, b))
            crops_args.append((fname, img.urls[mirror][b], *coords))

    out_dir = os.path.abspath(os.path.expanduser(out_dir))
    os.makedirs(out_dir, exist_ok=True)
    print('Downloading {} crops ({} images with {} bands)...'.format(len(crops_args),
                                                                     len(imgs),
                                                                     len(bands)),
          end=' ')
    parallel.run_calls(utils.crop_with_gdal_translate, crops_args,
                       extra_args=('UInt16',), pool_type='threads',
                       nb_workers=parallel_downloads)


def bands_files_are_valid(img, bands, api, directory):
    """
    Check if all bands images files are valid.
    """
    filenames = ['{}_band_{}.tif'.format(img.filename, b) for b in bands]
    paths = [os.path.join(directory, f) for f in filenames]
    return all(utils.is_valid(p) for p in paths)


def is_image_cloudy(img, aoi, mirror, p=0.5):
    """
    Tell if the given area of interest is covered by clouds in a given image.

    The location is considered covered if a fraction larger than p of its surface is
    labeled as clouds in the sentinel-2 gml cloud masks.

    Args:
        img (image object): Sentinel-2 image metadata
        aoi (geojson.Polygon): area of interest
        mirror (string): either 'gcloud' or 'aws'
        p (float): fraction threshold

    Return:
        boolean (True if the image is cloudy, False otherwise)
    """
    url = img.urls[mirror]['cloud_mask']

    if mirror == 'gcloud':
        gml_content = requests.get(url).text
    else:
        bucket, *key = url.replace('s3://', '').split('/')
        f = boto3.client('s3').get_object(Bucket=bucket, Key='/'.join(key),
                                          RequestPayer='requester')['Body']
        gml_content = f.read()

    clouds = []
    soup = bs4.BeautifulSoup(gml_content, 'xml')
    for polygon in soup.find_all('MaskFeature'):
        if polygon.maskType.text == 'OPAQUE':  # either OPAQUE or CIRRUS
            try:
                coords = list(map(float, polygon.posList.text.split()))
                points = list(zip(coords[::2], coords[1::2]))
                clouds.append(shapely.geometry.Polygon(points))
            except IndexError:
                pass
    aoi_shape = shapely.geometry.shape(aoi)
    try:
        cloudy = shapely.geometry.MultiPolygon(clouds).intersection(aoi_shape)
        return cloudy.area > (p * aoi_shape.area)
    except shapely.geos.TopologicalError:
        return False


def read_cloud_masks(aoi, imgs, bands, mirror, parallel_downloads, p=0.5,
                     out_dir=''):
    """
    Read Sentinel-2 GML cloud masks and intersects them with the input aoi.

    Args:
        aoi (geojson.Polygon): area of interest
        imgs (list): list of images
        bands (list): list of bands
        mirror (str): either 'aws' or 'gcloud'
        parallel_downloads (int): number of parallel gml files downloads
        p (float): cloud area threshold above which our aoi is said to be
            'cloudy' in the current image
    """
    print('Reading {} cloud masks...'.format(len(imgs)), end=' ')
    cloudy = parallel.run_calls(is_image_cloudy, imgs,
                                extra_args=(utils.geojson_lonlat_to_utm(aoi), mirror, p),
                                pool_type='threads',
                                nb_workers=parallel_downloads, verbose=True)
    print('{} cloudy images out of {}'.format(sum(cloudy), len(imgs)))

    for img, cloud in zip(imgs, cloudy):
        if cloud:
            out_dir = os.path.abspath(os.path.expanduser(out_dir))
            os.makedirs(os.path.join(out_dir, 'cloudy'), exist_ok=True)
            for b in bands:
                f = '{}_band_{}.tif'.format(img.filename, b)
                shutil.move(os.path.join(out_dir, f),
                            os.path.join(out_dir, 'cloudy', f))


def get_time_series(aoi, start_date=None, end_date=None, bands=['B04'],
                    out_dir='', api='devseed', mirror='gcloud',
                    product_type=None, cloud_masks=False,
                    parallel_downloads=multiprocessing.cpu_count()):
    """
    Main function: crop and download a time series of Sentinel-2 images.

    Args:
        aoi (geojson.Polygon): area of interest
        start_date (datetime.datetime, optional): start of the time range
        end_date (datetime.datetime, optional): end of the time range
        bands (list, optional): list of bands
        out_dir (str, optional): path where to store the downloaded crops
        api (str, optional): either devseed (default), scihub, planet or gcloud
        mirror (str, optional): either 'aws' or 'gcloud'
        product_type (str, optional): either 'L1C' or 'L2A'
        cloud_masks (bool, optional): if True, cloud masks are downloaded and
            cloudy images are discarded
        parallel_downloads (int): number of parallel gml files downloads
    """
    # check access to the selected search api and download mirror
    check_args(api, mirror, product_type)

    # default date range
    if end_date is None:
        end_date = datetime.date.today()
    if start_date is None:
        start_date = end_date - datetime.timedelta(91)  # 3 months

    # list available images
    images = search(aoi, start_date, end_date,
                    product_type=product_type, api=api)

    # download crops
    download(images, bands, aoi, mirror, out_dir, parallel_downloads)

    # discard images that failed to download
    images = [i for i in images if bands_files_are_valid(i, bands, api, out_dir)]

    # embed all metadata as GeoTIFF tags in the image files
    for img in images:
        metadata = vars(img)
        metadata['downloaded_by'] = 'TSD on {}'.format(datetime.datetime.now().isoformat())

        for b in bands:
            filepath = os.path.join(out_dir, '{}_band_{}.tif'.format(img.filename, b))
            utils.set_geotif_metadata_items(filepath, metadata)

    if cloud_masks:  # discard images that are totally covered by clouds
        read_cloud_masks(aoi, images, bands, mirror, parallel_downloads,
                         out_dir=out_dir)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description=('Automatic download and crop '
                                                  'of Sentinel-2 images'))
    parser.add_argument('--geom', type=utils.valid_geojson,
                        help=('path to geojson file'))
    parser.add_argument('--lat', type=utils.valid_lat,
                        help=('latitude of the center of the rectangle AOI'))
    parser.add_argument('--lon', type=utils.valid_lon,
                        help=('longitude of the center of the rectangle AOI'))
    parser.add_argument('-w', '--width', type=int, default=5000,
                        help='width of the AOI (m), default 5000 m')
    parser.add_argument('-l', '--height', type=int, default=5000,
                        help='height of the AOI (m), default 5000 m')
    parser.add_argument('-s', '--start-date', type=utils.valid_datetime,
                        help='start date, YYYY-MM-DD')
    parser.add_argument('-e', '--end-date', type=utils.valid_datetime,
                        help='end date, YYYY-MM-DD')
    parser.add_argument('-b', '--band', nargs='*', default=['B04'],
                        choices=ALL_BANDS + ['all'], metavar='',
                        help=('space separated list of spectral bands to'
                              ' download. Default is B04 (red). Allowed values'
                              ' are {}'.format(', '.join(ALL_BANDS))))
    parser.add_argument('-o', '--outdir', type=str, help=('path to save the '
                                                          'images'), default='')
    parser.add_argument('--api', type=str, choices=['devseed', 'planet', 'scihub', 'gcloud'],
                        default='devseed', help='search API')
    parser.add_argument('--mirror', type=str, choices=['aws', 'gcloud'],
                        default='gcloud', help='download mirror')
    parser.add_argument(
        '--product-type', choices=['L1C', 'L2A'], help='type of image')
    parser.add_argument('--parallel-downloads', type=int,
                        default=multiprocessing.cpu_count(),
                        help='max number of parallel crops downloads')
    parser.add_argument('--cloud-masks',  action='store_true',
                        help=('download cloud masks crops from provided GML files'))
    args = parser.parse_args()

    if 'all' in args.band:
        args.band = ALL_BANDS

    if args.geom and (args.lat or args.lon):
        parser.error('--geom and {--lat, --lon} are mutually exclusive')

    if not args.geom and (not args.lat or not args.lon):
        parser.error('either --geom or {--lat, --lon} must be defined')

    if args.geom:
        aoi = args.geom
    else:
        aoi = utils.geojson_geometry_object(args.lat, args.lon, args.width,
                                            args.height)
    get_time_series(aoi, start_date=args.start_date, end_date=args.end_date,
                    bands=args.band, out_dir=args.outdir, api=args.api,
                    mirror=args.mirror,
                    product_type=args.product_type,
                    cloud_masks=args.cloud_masks,
                    parallel_downloads=args.parallel_downloads)



# Copyright (C) 2017, Carlo de Franchis <carlo.de-franchis@m4x.org>

from __future__ import print_function
import os
import sys
import multiprocessing
import multiprocessing.pool


def show_progress(a):
    """
    Callback for the run_calls function: print nb of calls that are done.

    Args:
        a: useless argument, but since this function is used as a callback by
           apply_async, it has to take one argument.
    """
    show_progress.counter += 1
    status = '{:{fill}{width}} / {}'.format(show_progress.counter,
                                            show_progress.total,
                                            fill='',
                                            width=len(str(show_progress.total)))
    if show_progress.counter < show_progress.total:
        status += chr(8) * len(status)
    else:
        status += '\n'
    sys.stdout.write(status)
    sys.stdout.flush()


def run_calls(fun, list_of_args, extra_args=(), pool_type='processes',
              nb_workers=multiprocessing.cpu_count(), timeout=60, verbose=True,
              initializer=None, initargs=None):
    """
    Run a function several times in parallel with different inputs.

    Args:
        fun: function to be called several times in parallel.
        list_of_args: list of (first positional) arguments passed to fun, one
            per call
        extra_args: tuple containing extra arguments to be passed to fun
            (same value for all calls)
        pool_type: either 'processes' or 'threads'
        nb_workers: number of calls run simultaneously
        timeout: number of seconds allowed per function call
        verbose: either True (show the amount of computed calls) or False
        initializer, initargs (optional): if initializer is not None then each
            worker process will call initializer(*initargs) when it starts

    Return:
        list of outputs
    """
    if pool_type == 'processes':
        pool = multiprocessing.Pool(nb_workers, initializer, initargs)
    elif pool_type == 'threads':
        pool = multiprocessing.pool.ThreadPool(nb_workers)
    else:
        print('ERROR: unknow pool_type "{}"'.format(pool_type))

    results = []
    outputs = []
    if verbose:
        show_progress.counter = 0
        show_progress.total = len(list_of_args)
    for x in list_of_args:
        if type(x) == tuple:
            args = x + extra_args
        else:
            args = (x,) + extra_args
        results.append(pool.apply_async(fun, args=args,
                                        callback=show_progress if verbose else None))

    for r in results:
        try:
            outputs.append(r.get(timeout))
        except multiprocessing.TimeoutError:
            print("Timeout while running %s" % str(r), file=sys.stderr)
            outputs.append(None)
        except KeyboardInterrupt:
            pool.terminate()
            sys.exit(1)

    pool.close()
    pool.join()
    return outputs



import os
from codecs import open
from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))

package = "tsd"

about = {}
with open(os.path.join(here, package, "__about__.py"), "r", "utf-8") as f:
    exec(f.read(), about)

def readme():
    with open(os.path.join(here, 'README.md'), 'r', 'utf-8') as f:
        return f.read()

requirements = ['area',
                'boto3',
                'bs4',
                'click',
                'future',
                'geojson',
                'google-auth',
                'lxml',
                'mgrs',
                'numpy>=1.12',
                'pandas',
                'planet',
                'pyproj',
                'python-dateutil',
                'rasterio[s3]>=1.0',
                'requests',
                'sat-search>=0.2',
                'shapely',
                'tqdm',
                'utm',
                'sentinelhub']


setup(name=about["__title__"],
      version=about["__version__"],
      description=about["__description__"],
      long_description=readme(),
      long_description_content_type='text/markdown',
      url=about["__url__"],
      author=about["__author__"],
      author_email=about["__author_email__"],
      packages=[package],
      package_data={'': ['s2_mgrs_grid.txt']},
      include_package_data=True,
      install_requires=requirements,
      python_requires=">=3.5")



import argparse
import requests
import xmltodict


MUNDI_SEARCH_URL = 'https://catalog-browse.default.mundiwebservices.com/acdc/catalog/proxy/search'


def mundi_download_url_for_given_safe(safe_title):
    """
    Use Mundi search API to get the url of a given SAFE.

    Args:
        safe_title (string): title of a safe (without the .SAFE extension)

    Return:
        url to that SAFE zip file hosted at Mundi
    """
    r = requests.get('{}/Sentinel1/opensearch?uid={}'.format(MUNDI_SEARCH_URL, safe_title))
    if r.ok:
        d = xmltodict.parse(r.text)
        return d['feed']['entry']['link'][1]['@href']
    else:
        r.raise_for_status()


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Search of SAFE files on Mundi.')
    parser.add_argument('safename', help=('title of the safe'))
    args = parser.parse_args()
    print(mundi_download_url_for_given_safe(args.safename))



import datetime

from tsd import utils
from tsd import search_devseed

aoi = utils.geojson_geometry_object(29.9793, 31.1346, 5000, 5000)
results = search_devseed.search(aoi,
                                start_date=datetime.datetime(2019, 1, 15),
                                end_date=datetime.datetime(2019, 1, 30),
                                satellite='Sentinel-2')

expected_titles = ['S2B_MSIL1C_20190129T083209_N0207_R021_T36RUU_20190129T103220',
                   'S2A_MSIL1C_20190124T083231_N0207_R021_T36RUU_20190124T095836',
                   'S2B_MSIL1C_20190119T083259_N0207_R021_T36RUU_20190119T104924']
assert([r['properties']['sentinel:product_id'] for r in results] == expected_titles)



import dateutil

AWS_HTTPS_URL_L8 = 'https://landsat-pds.s3.amazonaws.com'
GCLOUD_URL_LANDSAT = 'https://storage.googleapis.com/gcp-public-data-landsat'
ALL_BANDS_LANDSAT = ['B{}'.format(i) for i in range(1, 12)] + ['BQA']


class LandsatGcloudParser:
    def __init__(self, img):
        self.meta = img
        self.urls = {'aws': {}, 'gcloud': {}}
        self._parse()
        self._build_gs_links()
        self.filename = '{}_{}_{}_{}'.format(self.date.date().isoformat(),
                                             self.satellite,
                                             self.sensor,
                                             self.scene_id)

    def _parse(self):
        d = self.meta.copy()
        self.scene_id = d['scene_id']
        self.satellite = d['spacecraft_id'].replace('ANDSAT_','')
        self.sensor = d['sensor_id'].replace('_', '')
        self.date = dateutil.parser.parse(d['sensing_time'], ignoretz=True)
        self.row = d['wrs_row']
        self.path = d['wrs_path']

    def _build_gs_links(self):
        base_url = self.meta['base_url']
        scene_id_bis = base_url.strip('/').split('/')[-1]
        for band in ALL_BANDS_LANDSAT:
            self.urls['gcloud'][band] = '{}/{}_{}.TIF'.format(base_url, scene_id_bis, band)


class LandsatDevSeedParser:
    def __init__(self, img):
        #self.meta = img
        p = img['properties']
        self.scene_id = p['landsat:scene_id']
        self.product_id = p['landsat:product_id']
        self.satellite = p['eo:platform'].replace('andsat-','').upper()
        self.sensor = p['eo:instrument'].replace('_', '')
        self.date = dateutil.parser.parse(p['datetime'], ignoretz=True)
        self.row = p['eo:row']
        self.path = p['eo:column']
        self.thumbnail = img['assets']['thumbnail']['href']
        self.cloud_cover = p['eo:cloud_cover']

        self.filename = '{}_{}_{}_{}'.format(self.date.date().isoformat(),
                                             self.satellite,
                                             self.sensor,
                                             self.scene_id)
        self.urls = {'aws': {}, 'gcloud': {}}

        if self.satellite == 'L8':
            for band in ALL_BANDS_LANDSAT:
                band_dico = img['assets'].get(band)
                if band_dico is not None:
                    self.urls['aws'][band] = band_dico['href']

        self._build_gs_links()

    def _build_gs_links(self):
        for band in ALL_BANDS_LANDSAT:
            sat_bis,*_,collec,_ = self.product_id.split('_')
            base_url = '{}/{}/{}/{}/{}/{}'.format(GCLOUD_URL_LANDSAT,sat_bis,collec,self.path,self.row, self.product_id)
            self.urls['gcloud'][band] = '{}/{}_{}.TIF'.format(base_url, self.product_id, band)
